'''
AVFoundation Camera
===================

Camera implementation using AVFoundation framework for OSX / iOS
'''

__all__ = ['CameraAVFoundation']

cdef extern from "camera_avfoundation_implem.h":
    ctypedef void *camera_t
    camera_t avf_camera_init(int index, int width, int height)
    void avf_camera_deinit(camera_t camera)
    bint avf_camera_update(camera_t camera)
    void avf_camera_start(camera_t camera)
    void avf_camera_stop(camera_t camera)
    void avf_camera_get_image(camera_t camera, int *width, int *height, int *rowsize, char **data)
    bint avf_camera_attempt_framerate_selection(camera_t camera, int fps)
    bint avf_camera_attempt_capture_preset(camera_t camera, char* preset)
    bint avf_camera_attempt_start_metadata_analysis(camera_t camera)
    void avf_camera_get_metadata(camera_t camera, char **metatype, char **data)
    bint avf_camera_have_new_metadata(camera_t camera);
    bint avf_camera_set_video_orientation(camera_t camera, int orientation)
    int avf_camera_get_device_orientation()
    void avf_camera_change_input(camera_t camera, int _cameraNum)
    void avf_camera_zoom_level(camera_t camera, float zoomLevel)
    char *avf_camera_documents_directory()
    void avf_camera_save_pixels(camera_t camera, unsigned char *pixels, int width, int height, char *path, float quality)


from kivy.logger import Logger
from kivy.clock import Clock
from kivy.graphics.texture import Texture
from kivy.core.camera import CameraBase
from kivy.utils import platform
from cython cimport view as cyview


cdef class _AVStorage:
    cdef camera_t camera

    def __cinit__(self):
        self.camera = NULL


class CameraAVFoundation(CameraBase):
    '''Implementation of CameraBase using AVFoundation
    '''

    def __init__(self, **kwargs):
        self._storage = _AVStorage()
        self._update_ev = None
        self._metadata_callback = None
        self._framerate = 30
        super(CameraAVFoundation, self).__init__(**kwargs)

    def __del__(self):
        self._release_camera()

    def init_camera(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        storage.camera = avf_camera_init(
            self._index, self.resolution[0], self.resolution[1])

    def _release_camera(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        if storage.camera != NULL:
            avf_camera_deinit(storage.camera)
            storage.camera = NULL

    @property
    def _scheduled_rate(self):
        # We're going 4 times faster the framerate to avoid frame skipping.
        return 1 / (self._framerate * 4)

    def _update(self, dt):
        cdef _AVStorage storage = <_AVStorage>self._storage
        cdef int width, height, rowsize
        cdef char *data
        cdef char *metadata_type
        cdef char *metadata_data
        cdef cyview.array cyarr

        if self.stopped:
            return

        if not avf_camera_update(storage.camera):
            return

        avf_camera_get_image(storage.camera,
            &width, &height, &rowsize, &data)

        if data == NULL:
            return

        cyarr = cyview.array(
            shape=(rowsize * height,),
            itemsize=sizeof(char),
            format="B",
            mode="c",
            allocate_buffer=False,
        )
        cyarr.data = data

        self._resolution = (width, height)
        
        if self._texture is None or self._texture.size != self._resolution:
            if platform == 'ios':
                self._texture = Texture.create(self._resolution, colorfmt='bgra')
            else:
                self._texture = Texture.create(self._resolution)
            self._texture.flip_vertical()
            self.dispatch('on_load')

        self._format = 'bgra'
        self._texture.blit_buffer(cyarr, colorfmt=self._format)
        self._copy_to_gpu()
        if self._metadata_callback:
            if avf_camera_have_new_metadata(storage.camera):
                avf_camera_get_metadata(storage.camera, &metadata_type, &metadata_data)
                self._metadata_callback(metadata_type, metadata_data)
    
    def _copy_to_gpu(self):
        self.dispatch('on_texture')

    def start(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        super(CameraAVFoundation, self).start()
        if self._update_ev is not None:
            self._update_ev.cancel()
        self._update_ev = Clock.schedule_interval(self._update, self._scheduled_rate)
        avf_camera_start(storage.camera)

    def stop(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        super(CameraAVFoundation, self).stop()
        if self._update_ev is not None:
            self._update_ev.cancel()
            self._update_ev = None
        avf_camera_stop(storage.camera)

    def set_framerate(self, framerate):
        cdef _AVStorage storage = <_AVStorage>self._storage
        if avf_camera_attempt_framerate_selection(storage.camera, framerate):
            if self._update_ev is not None:
                self._update_ev.cancel()
            self._framerate = framerate
            self._update_ev = Clock.schedule_interval(self._update, self._scheduled_rate)

    def set_preset(self, preset):
        cdef _AVStorage storage = <_AVStorage>self._storage
        avf_camera_attempt_capture_preset(storage.camera, preset)

    def set_video_orientation(self, orientation):
        cdef _AVStorage storage = <_AVStorage>self._storage
        avf_camera_set_video_orientation(storage.camera, orientation)

    def start_metadata_analysis(self, callback=None):
        cdef _AVStorage storage = <_AVStorage>self._storage
        self._metadata_callback = callback
        avf_camera_attempt_start_metadata_analysis(storage.camera)
        
    def get_device_orientation(self):
        # iOS only
        return avf_camera_get_device_orientation()

    def change_camera_input(self, index):
        # iOS only
        cdef _AVStorage storage = <_AVStorage>self._storage
        avf_camera_change_input(storage.camera, index)

    def zoom_level(self, level):
        # iOS only
        cdef _AVStorage storage = <_AVStorage>self._storage
        avf_camera_zoom_level(storage.camera, level)

    def get_app_documents_directory(self):
        # iOS only
        return str(avf_camera_documents_directory().decode('utf-8'))

    def save_texture(self, texture, filepath = '', quality = 0.9):
        # iOS only
        # With texture argument only: Save texture to iOS Photos App
        # With texture and filepath arguments: Save texture as jpg; filepath root must be app documents directory,
        #     sub-directories in filepath must exist, filepath last element is filename.jpg
        cdef _AVStorage storage = <_AVStorage>self._storage
        avf_camera_save_pixels(storage.camera, bytearray(texture.pixels), int(texture.width), int(texture.height),
                               filepath.encode('utf-8'), quality)




from libc.stdlib cimport malloc, free
from .bcm cimport DISPMANX_ELEMENT_HANDLE_T, ElementHandle
cimport kivy.lib.vidcore_lite.bcm as bcm
import kivy.lib.vidcore_lite.bcm as bcm

cdef extern from "EGL/egl.h":
    ctypedef int EGLint ###maybe wrong
    ctypedef unsigned int EGLBoolean
    ctypedef unsigned int EGLenum
    ctypedef void *EGLNativeDisplayType
    ctypedef void *EGLNativePixmapType
    ctypedef void *EGLNativeWindowType
    ctypedef void *EGLConfig
    ctypedef void *EGLContext
    ctypedef void *EGLDisplay
    ctypedef void *EGLSurface
    ctypedef void *EGLClientBuffer

    ctypedef struct EGL_DISPMANX_WINDOW_T:
        DISPMANX_ELEMENT_HANDLE_T element
        int width   #/* This is necessary because dispmanx elements are not queryable. */
        int height

    EGLint eglGetError()
    EGLDisplay eglGetDisplay(EGLNativeDisplayType display_id)
    EGLBoolean eglInitialize(EGLDisplay dpy, EGLint *major, EGLint *minor)
    EGLBoolean eglTerminate(EGLDisplay dpy)
    char * eglQueryString(EGLDisplay dpy, EGLint name)
    EGLBoolean eglGetConfigs(EGLDisplay dpy, EGLConfig *configs,
             EGLint config_size, EGLint *num_config)
    EGLBoolean eglChooseConfig(EGLDisplay dpy, EGLint *attrib_list,
               EGLConfig *configs, EGLint config_size,
               EGLint *num_config)
    EGLBoolean eglGetConfigAttrib(EGLDisplay dpy, EGLConfig config,
                  EGLint attribute, EGLint *value)
    EGLSurface eglCreateWindowSurface(EGLDisplay dpy, EGLConfig config,
                  EGLNativeWindowType win,
                  EGLint *attrib_list)
    EGLSurface eglCreatePbufferSurface(EGLDisplay dpy, EGLConfig config,
                   EGLint *attrib_list)
    EGLSurface eglCreatePixmapSurface(EGLDisplay dpy, EGLConfig config,
                  EGLNativePixmapType pixmap,
                  EGLint *attrib_list)
    EGLBoolean eglDestroySurface(EGLDisplay dpy, EGLSurface surface)
    EGLBoolean eglQuerySurface(EGLDisplay dpy, EGLSurface surface,
               EGLint attribute, EGLint *value)
    EGLBoolean eglBindAPI(EGLenum api)
    EGLenum eglQueryAPI()
    EGLBoolean eglWaitClient()
    EGLBoolean eglReleaseThread()
    EGLSurface eglCreatePbufferFromClientBuffer(
          EGLDisplay dpy, EGLenum buftype, EGLClientBuffer buffer,
          EGLConfig config, EGLint *attrib_list)
    EGLBoolean eglSurfaceAttrib(EGLDisplay dpy, EGLSurface surface,
                EGLint attribute, EGLint value)
    EGLBoolean eglBindTexImage(EGLDisplay dpy, EGLSurface surface, EGLint buffer)
    EGLBoolean eglReleaseTexImage(EGLDisplay dpy, EGLSurface surface, EGLint buffer)
    EGLBoolean eglSwapInterval(EGLDisplay dpy, EGLint interval)
    EGLContext eglCreateContext(EGLDisplay dpy, EGLConfig config,
                EGLContext share_context,
                EGLint *attrib_list)
    EGLBoolean eglDestroyContext(EGLDisplay dpy, EGLContext ctx)
    EGLBoolean eglMakeCurrent(EGLDisplay dpy, EGLSurface draw,
              EGLSurface read, EGLContext ctx)
    EGLContext eglGetCurrentContext()
    EGLSurface eglGetCurrentSurface(EGLint readdraw)
    EGLDisplay eglGetCurrentDisplay()
    EGLBoolean eglQueryContext(EGLDisplay dpy, EGLContext ctx,
               EGLint attribute, EGLint *value)
    EGLBoolean eglWaitGL()
    EGLBoolean eglWaitNative(EGLint engine)
    EGLBoolean eglSwapBuffers(EGLDisplay dpy, EGLSurface surface)
    EGLBoolean eglCopyBuffers(EGLDisplay dpy, EGLSurface surface,
              EGLNativePixmapType target)

class _constants:
    EGL_VERSION_1_0 = 1 #
    EGL_VERSION_1_1 = 1 #
    EGL_VERSION_1_2 = 1 #
    EGL_VERSION_1_3 = 1 #
    EGL_VERSION_1_4 = 1 #

    ''' EGL Enumerants. Bitmasks and other exceptional cases aside, most
     * enums are assigned unique values starting at 0x3000.
     '''

    ''' EGL aliases '''
    EGL_FALSE = 0 #
    EGL_TRUE = 1 #

    ''' Out-of-band handle values '''
    EGL_DEFAULT_DISPLAY = 0 #
    EGL_NO_CONTEXT = 0 #
    EGL_NO_DISPLAY = 0 #
    EGL_NO_SURFACE = 0 #

    ''' Out-of-band attribute value '''
    EGL_DONT_CARE = -1 #

    ''' Errors / GetError return values '''
    EGL_SUCCESS = 0x3000 #
    EGL_NOT_INITIALIZED = 0x3001 #
    EGL_BAD_ACCESS = 0x3002 #
    EGL_BAD_ALLOC = 0x3003 #
    EGL_BAD_ATTRIBUTE = 0x3004 #
    EGL_BAD_CONFIG = 0x3005 #
    EGL_BAD_CONTEXT = 0x3006 #
    EGL_BAD_CURRENT_SURFACE = 0x3007 #
    EGL_BAD_DISPLAY = 0x3008 #
    EGL_BAD_MATCH = 0x3009 #
    EGL_BAD_NATIVE_PIXMAP = 0x300A #
    EGL_BAD_NATIVE_WINDOW = 0x300B #
    EGL_BAD_PARAMETER = 0x300C #
    EGL_BAD_SURFACE = 0x300D #
    EGL_CONTEXT_LOST = 0x300E # ''' EGL 1.1 - IMG_power_management '''

    ''' Reserved 0x300F-0x301F for additional errors '''

    ''' Reserved 0x3041-0x304F for additional config attributes '''

    ''' Config attribute values '''
    EGL_SLOW_CONFIG = 0x3050 # ''' EGL_CONFIG_CAVEAT value '''
    EGL_NON_CONFORMANT_CONFIG = 0x3051 # ''' EGL_CONFIG_CAVEAT value '''
    EGL_TRANSPARENT_RGB = 0x3052 # ''' EGL_TRANSPARENT_TYPE value '''
    EGL_RGB_BUFFER = 0x308E # ''' EGL_COLOR_BUFFER_TYPE value '''
    EGL_LUMINANCE_BUFFER = 0x308F # ''' EGL_COLOR_BUFFER_TYPE value '''

    ''' More config attribute values, for EGL_TEXTURE_FORMAT '''
    EGL_NO_TEXTURE = 0x305C #
    EGL_TEXTURE_RGB = 0x305D #
    EGL_TEXTURE_RGBA = 0x305E #
    EGL_TEXTURE_2D = 0x305F #

    ''' Config attribute mask bits '''
    EGL_PBUFFER_BIT = 0x0001 # ''' EGL_SURFACE_TYPE mask bits '''
    EGL_PIXMAP_BIT = 0x0002 # ''' EGL_SURFACE_TYPE mask bits '''
    EGL_WINDOW_BIT = 0x0004 # ''' EGL_SURFACE_TYPE mask bits '''
    EGL_VG_COLORSPACE_LINEAR_BIT = 0x0020 # ''' EGL_SURFACE_TYPE mask bits '''
    EGL_VG_ALPHA_FORMAT_PRE_BIT = 0x0040 # ''' EGL_SURFACE_TYPE mask bits '''
    EGL_MULTISAMPLE_RESOLVE_BOX_BIT = 0x0200 # ''' EGL_SURFACE_TYPE mask bits '''
    EGL_SWAP_BEHAVIOR_PRESERVED_BIT = 0x0400 # ''' EGL_SURFACE_TYPE mask bits '''

    EGL_OPENGL_ES_BIT = 0x0001 # ''' EGL_RENDERABLE_TYPE mask bits '''
    EGL_OPENVG_BIT = 0x0002 # ''' EGL_RENDERABLE_TYPE mask bits '''
    EGL_OPENGL_ES2_BIT = 0x0004 # ''' EGL_RENDERABLE_TYPE mask bits '''
    EGL_OPENGL_BIT = 0x0008 # ''' EGL_RENDERABLE_TYPE mask bits '''

    ''' QueryString targets '''
    EGL_VENDOR = 0x3053 #
    EGL_VERSION = 0x3054 #
    EGL_EXTENSIONS = 0x3055 #
    EGL_CLIENT_APIS = 0x308D #

    ''' QuerySurface / SurfaceAttrib / CreatePbufferSurface targets '''
    EGL_HEIGHT = 0x3056 #
    EGL_WIDTH = 0x3057 #
    EGL_LARGEST_PBUFFER = 0x3058 #
    EGL_TEXTURE_FORMAT = 0x3080 #
    EGL_TEXTURE_TARGET = 0x3081 #
    EGL_MIPMAP_TEXTURE = 0x3082 #
    EGL_MIPMAP_LEVEL = 0x3083 #
    EGL_RENDER_BUFFER = 0x3086 #
    EGL_VG_COLORSPACE = 0x3087 #
    EGL_VG_ALPHA_FORMAT = 0x3088 #
    EGL_HORIZONTAL_RESOLUTION = 0x3090 #
    EGL_VERTICAL_RESOLUTION = 0x3091 #
    EGL_PIXEL_ASPECT_RATIO = 0x3092 #
    EGL_SWAP_BEHAVIOR = 0x3093 #
    EGL_MULTISAMPLE_RESOLVE = 0x3099 #

    ''' EGL_RENDER_BUFFER values / BindTexImage / ReleaseTexImage buffer targets '''
    EGL_BACK_BUFFER = 0x3084 #
    EGL_SINGLE_BUFFER = 0x3085 #

    ''' OpenVG color spaces '''
    EGL_VG_COLORSPACE_sRGB = 0x3089 # ''' EGL_VG_COLORSPACE value '''
    EGL_VG_COLORSPACE_LINEAR = 0x308A # ''' EGL_VG_COLORSPACE value '''

    ''' OpenVG alpha formats '''
    EGL_VG_ALPHA_FORMAT_NONPRE = 0x308B # ''' EGL_ALPHA_FORMAT value '''
    EGL_VG_ALPHA_FORMAT_PRE = 0x308C # ''' EGL_ALPHA_FORMAT value '''

    ''' Constant scale factor by which fractional display resolutions &
     * aspect ratio are scaled when queried as integer values.
     '''
    EGL_DISPLAY_SCALING = 10000 #

    ''' Unknown display resolution/aspect ratio '''
    EGL_UNKNOWN = -1 #

    ''' Back buffer swap behaviors '''
    EGL_BUFFER_PRESERVED = 0x3094 # ''' EGL_SWAP_BEHAVIOR value '''
    EGL_BUFFER_DESTROYED = 0x3095 # ''' EGL_SWAP_BEHAVIOR value '''

    ''' CreatePbufferFromClientBuffer buffer types '''
    EGL_OPENVG_IMAGE = 0x3096 #

    ''' QueryContext targets '''
    EGL_CONTEXT_CLIENT_TYPE = 0x3097 #

    ''' CreateContext attributes '''
    EGL_CONTEXT_CLIENT_VERSION = 0x3098 #

    ''' Multisample resolution behaviors '''
    EGL_MULTISAMPLE_RESOLVE_DEFAULT = 0x309A # ''' EGL_MULTISAMPLE_RESOLVE value '''
    EGL_MULTISAMPLE_RESOLVE_BOX = 0x309B # ''' EGL_MULTISAMPLE_RESOLVE value '''

    ''' GetCurrentSurface targets '''
    EGL_DRAW = 0x3059 #
    EGL_READ = 0x305A #

    ''' WaitNative engines '''
    EGL_CORE_NATIVE_ENGINE = 0x305B #

    ''' EGL 1.2 tokens renamed for consistency in EGL 1.3 '''
    EGL_COLORSPACE = EGL_VG_COLORSPACE #
    EGL_ALPHA_FORMAT = EGL_VG_ALPHA_FORMAT #
    EGL_COLORSPACE_sRGB = EGL_VG_COLORSPACE_sRGB #
    EGL_COLORSPACE_LINEAR = EGL_VG_COLORSPACE_LINEAR #
    EGL_ALPHA_FORMAT_NONPRE = EGL_VG_ALPHA_FORMAT_NONPRE #
    EGL_ALPHA_FORMAT_PRE = EGL_VG_ALPHA_FORMAT_PRE #


    EGL_OPENGL_ES_API = 0x30A0
    EGL_OPENVG_API	=	0x30A1
    EGL_OPENGL_API	=	0x30A2

    EGL_BUFFER_SIZE	=		0x3020
    EGL_ALPHA_SIZE	=		0x3021
    EGL_BLUE_SIZE	=		0x3022
    EGL_GREEN_SIZE	=		0x3023
    EGL_RED_SIZE	=		0x3024
    EGL_DEPTH_SIZE	=		0x3025
    EGL_STENCIL_SIZE=		0x3026
    EGL_CONFIG_CAVEAT	=	0x3027
    EGL_CONFIG_ID		=	0x3028
    EGL_LEVEL	=		0x3029
    EGL_MAX_PBUFFER_HEIGHT	=	0x302A
    EGL_MAX_PBUFFER_PIXELS	=	0x302B
    EGL_MAX_PBUFFER_WIDTH	=	0x302C
    EGL_NATIVE_RENDERABLE	=	0x302D
    EGL_NATIVE_VISUAL_ID	=	0x302E
    EGL_NATIVE_VISUAL_TYPE	=	0x302F
    EGL_SAMPLES		=	0x3031
    EGL_SAMPLE_BUFFERS	=	0x3032
    EGL_SURFACE_TYPE	=	0x3033
    EGL_TRANSPARENT_TYPE	=	0x3034
    EGL_TRANSPARENT_BLUE_VALUE=	0x3035
    EGL_TRANSPARENT_GREEN_VALUE=	0x3036
    EGL_TRANSPARENT_RED_VALUE=	0x3037
    EGL_NONE	=		0x3038	#/* Attrib list terminator */
    EGL_BIND_TO_TEXTURE_RGB	=	0x3039
    EGL_BIND_TO_TEXTURE_RGBA=	0x303A
    EGL_MIN_SWAP_INTERVAL	=	0x303B
    EGL_MAX_SWAP_INTERVAL	=	0x303C
    EGL_LUMINANCE_SIZE	=	0x303D
    EGL_ALPHA_MASK_SIZE	=	0x303E
    EGL_COLOR_BUFFER_TYPE	=	0x303F
    EGL_RENDERABLE_TYPE	=	0x3040
    EGL_MATCH_NATIVE_PIXMAP	=	0x3041	#/* Pseudo-attribute (not queryable) */
    EGL_CONFORMANT	=		0x3042

EGL_FALSE = _constants.EGL_FALSE

global _context_reg
_context_reg = {}

cdef class Context:
    cdef EGLContext _eglcontext

global _display_reg
_display_reg = {}

cdef class Display:
    cdef EGLDisplay _egldisplay

global _surface_reg
_surface_reg = {}

cdef class Surface:
    cdef EGLSurface _eglsurface

global _config_reg
_config_reg = {}

cdef class Config:
    cdef EGLConfig _eglconfig


class EGLError(Exception):
    codes = {
        0x3000 : 'EGL_SUCCESS'			,
        0x3001 : 'EGL_NOT_INITIALIZED',
        0x3002 : 'EGL_BAD_ACCESS'		,
        0x3003 : 'EGL_BAD_ALLOC'		,
        0x3004 : 'EGL_BAD_ATTRIBUTE'	,
        0x3005 : 'EGL_BAD_CONFIG'		,
        0x3006 : 'EGL_BAD_CONTEXT'		,
        0x3007 : 'EGL_BAD_CURRENT_SURFACE',
        0x3008 : 'EGL_BAD_DISPLAY'		,
        0x3009 : 'EGL_BAD_MATCH'			,
        0x300A : 'EGL_BAD_NATIVE_PIXMAP',
        0x300B : 'EGL_BAD_NATIVE_WINDOW',
        0x300C : 'EGL_BAD_PARAMETER'	,
        0x300D : 'EGL_BAD_SURFACE'		,
        0x300E : 'EGL_CONTEXT_LOST'		,
        }


cdef class NativeWindow:
    cdef EGL_DISPMANX_WINDOW_T _window

    def __cinit__(self, ElementHandle element, int width, int height):
        self._window.element = element._handle
        self._window.width = width
        self._window.height = height


def raise_egl_error():
    err_code = getError()
    raise EGLError("%s (code x%x)"%(EGLError.codes[err_code], err_code))

def getError():
    return int(eglGetError())

def BindAPI(EGLenum api):
    cdef:
        EGLBoolean ret
    ret = eglBindAPI(api)
    return bool(ret)

def GetDisplay(unsigned int display_id):
    cdef:
        EGLDisplay display
        Display py_display
    display = eglGetDisplay(<EGLNativeDisplayType>display_id)
    if display == <EGLDisplay>0:
        raise EGLError("No display available")
    py_display = Display()
    py_display._egldisplay = display
    _display_reg[<int>display] = py_display
    return py_display

def Initialise(Display dpy):
    cdef:
        EGLint major
        EGLint minor
        EGLBoolean ret
    ret = eglInitialize(dpy._egldisplay, &major, &minor)
    if ret == EGL_FALSE:
        raise_egl_error()
    return (int(major), int(minor))

def Terminate(Display dpy):
    if eglTerminate(dpy._egldisplay) == EGL_FALSE:
        raise_egl_error()

def QueryString(Display dpy, EGLint name):
    cdef char *data
    data = eglQueryString(dpy._egldisplay, name)
    return data

def GetConfigs(Display dpy):
    cdef:
        EGLint numConfigs = -1
    if eglGetConfigs(dpy._egldisplay, NULL, 0, &numConfigs)==EGL_FALSE:
        raise_egl_error()
    return int(numConfigs)

### Not implemented the other calling method for this ###
#    EGLBoolean eglGetConfigs(EGLDisplay dpy, EGLConfig *configs,
#			 EGLint config_size, EGLint *num_config)

def ChooseConfig(Display dpy, list attrib_list, EGLint config_size):
    cdef:
        EGLConfig *configs
        EGLint num_config
        EGLint *attribs
        int i, n_attrib=len(attrib_list)

    attribs = <EGLint*>malloc(sizeof(EGLint)*n_attrib)
    configs = <EGLConfig*>malloc(sizeof(EGLConfig)*config_size)
    try:
        for i in xrange(n_attrib):
            attribs[i] = attrib_list[i]

        if eglChooseConfig(dpy._egldisplay, attribs,
                   configs, config_size, &num_config) == EGL_FALSE:
            raise_egl_error()
        chosen = []
        for i in xrange(num_config):
            cfg = Config()
            cfg._eglconfig = configs[i]
            chosen.append(cfg)
        return chosen
    finally:
        free(attribs)
        free(configs)

def GetConfigAttrib(Display dpy, Config config, EGLint attribute):
    cdef EGLint value
    if eglGetConfigAttrib(dpy._egldisplay, config._eglconfig,
                  attribute, &value) == EGL_FALSE:
        raise_egl_error()
    return int(value)

def CreateWindowSurface(Display dpy, Config config, NativeWindow win,
                        list attrib_list=[]):
    cdef:
        EGLSurface surf
        EGLint *attribs=NULL
        int i, n_attrib=len(attrib_list)

    if n_attrib > 0:
        attribs = <EGLint*>malloc(sizeof(EGLint)*n_attrib)
        for i in xrange(n_attrib):
            attribs[i] = attrib_list[i]
    try:
        surf = eglCreateWindowSurface(dpy._egldisplay, config._eglconfig,
                      <EGLNativeWindowType>(&(win._window)), attribs) #FIXME
        if surf == <EGLSurface>0:
            raise_egl_error()
        py_surf = Surface()
        py_surf._eglsurface = surf
        _surface_reg[<int>surf] = py_surf
        return py_surf
    finally:
        free(attribs)


def CreatePbufferSurface(Display dpy, Config config, list attrib_list):
    cdef:
        EGLSurface surf
        EGLint *attribs
        int i, n_attrib=len(attrib_list)

    attribs = <EGLint*>malloc(sizeof(EGLint)*n_attrib)
    try:
        for i in xrange(n_attrib):
            attribs[i] = attrib_list[i]

        surf = eglCreatePbufferSurface(dpy._egldisplay, config._eglconfig,
                      attribs)
        if surf == <EGLSurface>0:
            raise_egl_error()
        py_surf = Surface()
        py_surf._eglsurface = surf
        _surface_reg[<int>surf] = py_surf
        return py_surf
    finally:
        free(attribs)

#def CreatePixmapSurface(Display dpy, Config config, pixmap, list attrib_list):
#    cdef:
#        EGLSurface surf
#        EGLint *attribs
#        int i, n_attrib=len(attrib_list)
#
#    attribs = <EGLint*>malloc(sizeof(EGLint)*n_attrib)
#    try:
#        for i in xrange(n_attrib):
#            attribs[i] = attrib_list[i]
#
#        surf = eglCreatePixmapSurface(dpy._egldisplay, config._eglconfig,
#                      pixmap, attribs)
#        if surf == <EGLSurface>0:
#            raise_egl_error()
#        py_surf = Surface()
#        py_surf._eglsurface = surf
#        _surface_rep[<int>surf] = py_surf
#        return py_surf
#    finally:
#        free(attribs)

def DestroySurface(Display dpy, Surface surf):
    if eglDestroySurface(dpy._egldisplay, surf._eglsurface) == EGL_FALSE:
        raise_egl_error()
    del _surface_reg[<int>(surf._eglsurface)]

def QuerySurface(Display dpy, Surface surf, EGLint attrib):
    cdef EGLint ret
    if eglQuerySurface(dpy._egldisplay, surf._eglsurface,
               attrib, &ret) == EGL_FALSE:
        raise_egl_error()
    return ret

def QueryAPI():
    return eglQueryAPI()

def WaitClient():
    if eglWaitClient() == EGL_FALSE:
        raise_egl_error()

def ReleaseThread():
    if eglReleaseThread() == EGL_FALSE:
        raise_egl_error()

####
#def CreatePbufferFromClientBuffer(Display dpy, EGLenum buftype, buf,
#                    Config config, list attrib_list):
#    FIXME
#    EGLSurface eglCreatePbufferFromClientBuffer(
#	      EGLDisplay dpy, EGLenum buftype, EGLClientBuffer buffer,
#	      EGLConfig config, EGLint *attrib_list)


def SurfaceAttrib(Display dpy, Surface surf, EGLint attrib, EGLint value):
    cdef EGLBoolean ret
    ret = eglSurfaceAttrib(dpy._egldisplay, surf._eglsurface,
                            attrib, value)
    if ret==EGL_FALSE: raise_egl_error()

def BindTexImage(Display dpy, Surface surf, EGLint buf):
    if eglBindTexImage(dpy._egldisplay, surf._eglsurface, buf) == EGL_FALSE:
        raise_egl_error()

def ReleaseTexImage(Display dpy, Surface surf, EGLint buf):
    if eglReleaseTexImage(dpy._egldisplay, surf._eglsurface, buf) == EGL_FALSE:
        raise_egl_error()

def SwapInterval(Display dpy, EGLint interval):
    if eglSwapInterval(dpy._egldisplay, interval) == EGL_FALSE:
        raise_egl_error()

def CreateContext(Display dpy, Config config, object share_ctx, list attrib_list=[]):
    cdef:
        EGLContext _ctx, _share_ctx
        EGLint *attribs=NULL
        int i, n_attrib=len(attrib_list)

    if share_ctx is None:
        _share_ctx = <EGLContext>0
    elif isinstance(share_ctx, Context):
        _share_ctx = (<Context>share_ctx)._eglcontext
    else:
        raise ValueError("3rd argument (share context) must be of type Context or None")

    if n_attrib > 0:
        attribs = <EGLint*>malloc(sizeof(EGLint)*n_attrib)
    try:
        for i in xrange(n_attrib):
            attribs[i] = attrib_list[i]

        _ctx = eglCreateContext(dpy._egldisplay, config._eglconfig,
                    _share_ctx, attribs)
        pyctx = Context()
        pyctx._eglcontext = _ctx
        _context_reg[<int>_ctx] = pyctx
        return pyctx
    finally:
        free(attribs)

def DestroyContext(Display dpy, Context ctx):
    if eglDestroyContext(dpy._egldisplay, ctx._eglcontext) == EGL_FALSE:
        raise_egl_error()
    del _context_reg[<int>(ctx._eglcontext)]

def MakeCurrent(Display dpy, Surface draw, Surface read, Context ctx):
    if eglMakeCurrent(dpy._egldisplay, draw._eglsurface,
              read._eglsurface, ctx._eglcontext) == EGL_FALSE:
        raise_egl_error()

def GetCurrentContext():
    cdef EGLContext ctx
    ctx = eglGetCurrentContext()
    return _context_reg[<int>ctx]

def GetCurrentSurface(EGLint readdraw):
    cdef EGLSurface surf
    surf = eglGetCurrentSurface(readdraw)
    return _surface_reg[<int>surf]

def GetCurrentDisplay():
    cdef EGLDisplay dpy
    dpy = eglGetCurrentDisplay()
    return _display_reg[<int>dpy]

def QueryContext(Display dpy, Context ctx, EGLint attrib):
    cdef EGLint value
    if eglQueryContext(dpy._egldisplay, ctx._eglcontext,
                        attrib, &value) == EGL_FALSE:
        raise_egl_error()
    return value

def WaitGL():
    if eglWaitGL() == EGL_FALSE:
        raise_egl_error()

def WaitNative(EGLint engine):
    if eglWaitNative(engine) == EGL_FALSE:
        raise_egl_error()

def SwapBuffers(Display dpy, Surface surf):
    if eglSwapBuffers(dpy._egldisplay, surf._eglsurface) == EGL_FALSE:
        raise_egl_error()


#### Don't know what a NativePixmapType is or how to make one ###
#def CopyBuffers(Display dpy, Surface surf, target):
#    EGLBoolean eglCopyBuffers(EGLDisplay dpy, EGLSurface surface,
#			  EGLNativePixmapType target)
def bcm_display_open(bcm.uint32_t device):
    cdef:
        bcm.DISPMANX_DISPLAY_HANDLE_T disp
        bcm.DisplayHandle D
    disp = bcm.vc_dispmanx_display_open( device )
    if disp == 0:
        raise bcm.BCMDisplayException("Couldn't open handle to display")
    D = bcm.DisplayHandle()
    D._handle = disp
    return D

def bcm_update_start(bcm.int32_t priority):
    cdef:
        bcm.DISPMANX_UPDATE_HANDLE_T hdl
        bcm.UpdateHandle U
    hdl = bcm.vc_dispmanx_update_start( priority )
    if hdl == 0:
        raise bcm.BCMDisplayException("Couldn't open handle to update-start")
    U = bcm.UpdateHandle()
    U._handle = hdl
    return U

def bcm_element_add(bcm.UpdateHandle update,
                bcm.DisplayHandle display,
                bcm.int32_t layer,
                bcm.Rect dest_rect,
                bcm.Rect src_rect):
    cdef:
        bcm.DISPMANX_ELEMENT_HANDLE_T elem
        bcm.ElementHandle E
    elem = bcm.vc_dispmanx_element_add (update._handle,
                                     display._handle,
                                     layer,
                                     &(dest_rect._vc_rect),
                                     0, #DISPMANX_RESOURCE_HANDLE_T src,
                                     &(src_rect._vc_rect),
                                     0, #DISPMANX_PROTECTION_T protection,
                                     <bcm.VC_DISPMANX_ALPHA_T *>0, #VC_DISPMANX_ALPHA_T *alpha,
                                     <bcm.DISPMANX_CLAMP_T *>0, #DISPMANX_CLAMP_T *clamp,
                                     <bcm.DISPMANX_TRANSFORM_T>0) #DISPMANX_TRANSFORM_T transform
    E = bcm.ElementHandle()
    E._handle = elem
    return E

def bcm_update_submit_sync(bcm.UpdateHandle update):
    return bcm.vc_dispmanx_update_submit_sync( update._handle )

def WinCreate2(NativeWindow nativewindow, bcm.DisplayHandle display,
                    bcm.UpdateHandle update,
                    bcm.Rect dst, bcm.Rect src, display_id=0):
    cdef:
        bcm.int32_t success = 0
        bcm.DISPMANX_ELEMENT_HANDLE_T dispman_element
        bcm.DISPMANX_DISPLAY_HANDLE_T dispman_display
        bcm.DISPMANX_UPDATE_HANDLE_T dispman_update
        #bcm.VC_RECT_T dst_rect
        #bcm.VC_RECT_T src_rect
        bcm.uint32_t display_width
        bcm.uint32_t display_height

    ## create an EGL window surface, passing context width/height
    success = bcm.c_get_display_size(display_id, ## /* LCD */
                                            &display_width,
                                            &display_height);
    if ( success < 0 ):
        raise RuntimeError("Couldn't get display size")

    dispman_display = display._handle
    dispman_update = update._handle
    #dispman_update = bcm.vc_dispmanx_update_start( 0 )

    dispman_element = bcm.vc_dispmanx_element_add ( dispman_update, dispman_display,
        0, ##/*layer*/,
        &(dst._vc_rect),
        <bcm.DISPMANX_RESOURCE_HANDLE_T>0, ##/*src*/,
        &(src._vc_rect),
        <bcm.DISPMANX_PROTECTION_T>0,
        <bcm.VC_DISPMANX_ALPHA_T *>0, ##/*alpha*/
        <bcm.DISPMANX_CLAMP_T *>0, ##/*clamp*/
        <bcm.DISPMANX_TRANSFORM_T>0) ##/*transform*/

    nativewindow._window.element = dispman_element
    nativewindow._window.width = dst._vc_rect.width
    nativewindow._window.height = dst._vc_rect.height
    bcm.vc_dispmanx_update_submit_sync( dispman_update )
    return True



#cython: c_string_type=unicode, c_string_encoding=utf8
'''
TODO:
    - ensure that we correctly check allocation
    - remove compat sdl usage (like SDL_SetAlpha must be replaced with sdl 1.3
      call, not 1.2)
'''

include '../../lib/sdl2.pxi'

from kivy.core.image import ImageData
from kivy.compat import PY2

cdef dict sdl2_cache = {}
cdef list sdl2_cache_order = []

cdef class _TTFContainer:
    cdef TTF_Font* font
    def __cinit__(self):
        self.font = NULL
    def __dealloc__(self):
        if self.font != NULL:
            TTF_CloseFont(self.font)
            self.font = NULL


cdef class _SurfaceContainer:
    cdef SDL_Surface* surface
    cdef int w, h

    def __cinit__(self, w, h):
        self.surface = NULL
        self.w = w
        self.h = h

    def __init__(self, w, h):
        # XXX check on OSX to see if little endian/big endian make a difference
        # here.
        self.surface = SDL_CreateRGBSurface(0,
            w, h, 32,
            0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)

    def __dealloc__(self):
        if self.surface != NULL:
            SDL_FreeSurface(self.surface)
            self.surface = NULL

    def render(self, container, text, x, y):
        cdef TTF_Font *font = _get_font(container)
        cdef SDL_Color c
        cdef SDL_Color oc
        cdef SDL_Surface *st
        cdef SDL_Surface *fgst
        cdef SDL_Rect r
        cdef SDL_Rect fgr
        cdef list color = list(container.options['color'])
        cdef list outline_color = list(container.options['outline_color'])
        outline_width = container.options['outline_width']
        if font == NULL:
            return

        c.a = oc.a = 255
        c.r = <int>(color[0] * 255)
        c.g = <int>(color[1] * 255)
        c.b = <int>(color[2] * 255)
        bytes_text = <bytes>text.encode('utf-8')

        hinting = container.options['font_hinting']
        if hinting == 'normal':
            if TTF_GetFontHinting(font) != TTF_HINTING_NORMAL:
                TTF_SetFontHinting(font, TTF_HINTING_NORMAL)
        elif hinting == 'light':
            if TTF_GetFontHinting(font) != TTF_HINTING_LIGHT:
                TTF_SetFontHinting(font, TTF_HINTING_LIGHT)
        elif hinting == 'mono':
            if TTF_GetFontHinting(font) != TTF_HINTING_MONO:
                TTF_SetFontHinting(font, TTF_HINTING_MONO)
        elif hinting is None:
            if TTF_GetFontHinting(font) != TTF_HINTING_NONE:
                TTF_SetFontHinting(font, TTF_HINTING_NONE)

        if container.options['font_kerning']:
            if TTF_GetFontKerning(font) == 0:
                TTF_SetFontKerning(font, 1)
        else:
            if TTF_GetFontKerning(font) != 0:
                TTF_SetFontKerning(font, 0)

        direction = container.options['font_direction']
        if direction == 'ltr':
            TTF_SetFontDirection(font, TTF_DIRECTION_LTR)
        elif direction == 'rtl':
            TTF_SetFontDirection(font, TTF_DIRECTION_RTL)
        elif direction == 'ttb':
            TTF_SetFontDirection(font, TTF_DIRECTION_TTB)
        elif direction == 'btt':
            TTF_SetFontDirection(font, TTF_DIRECTION_BTT)

        fontscript = container.options['font_script_name']
        TTF_SetFontScriptName(font, fontscript)

        if outline_width:
            TTF_SetFontOutline(font, outline_width)
            oc.r = <int>(outline_color[0] * 255)
            oc.g = <int>(outline_color[1] * 255)
            oc.b = <int>(outline_color[2] * 255)
            st = (
                TTF_RenderUTF8_Blended(font, <char *>bytes_text, oc)
                if container.options['font_blended']
                else TTF_RenderUTF8_Solid(font, <char *>bytes_text, oc)
                )
            TTF_SetFontOutline(font, 0)
        else:
            st = (
                TTF_RenderUTF8_Blended(font, <char *>bytes_text, c)
                if container.options['font_blended']
                else TTF_RenderUTF8_Solid(font, <char *>bytes_text, c)
                )
        if st == NULL:
            return
        if outline_width:
            fgst = (
                TTF_RenderUTF8_Blended(font, <char *>bytes_text, c)
                if container.options['font_blended']
                else TTF_RenderUTF8_Solid(font, <char *>bytes_text, c)
                )
            if fgst == NULL:
                SDL_FreeSurface(st)
                return
            fgr.x = outline_width
            fgr.y = outline_width
            fgr.w = fgst.w
            fgr.h = fgst.h
            SDL_SetSurfaceBlendMode(fgst, SDL_BLENDMODE_BLEND)
            SDL_BlitSurface(fgst, NULL, st, &fgr)
            SDL_FreeSurface(fgst)

        r.x = x
        r.y = y
        r.w = st.w
        r.h = st.h
        SDL_SetSurfaceAlphaMod(st, <int>(color[3] * 255))
        if container.options['line_height'] < 1:
            """
            We are using SDL_BLENDMODE_BLEND only when line_height < 1 as a workaround.
            Previously, We enabled SDL_BLENDMODE_BLEND also for text w/ line_height >= 1,
            but created an unexpected behavior (See PR #6507 and issue #6508).
            """
            SDL_SetSurfaceBlendMode(st, SDL_BLENDMODE_BLEND)
        else:
            SDL_SetSurfaceBlendMode(st, SDL_BLENDMODE_NONE)
        SDL_BlitSurface(st, NULL, self.surface, &r)
        SDL_FreeSurface(st)

    def get_data(self):
        cdef int datalen = self.surface.w * self.surface.h * 4
        cdef bytes pixels = (<char *>self.surface.pixels)[:datalen]
        data = ImageData(self.w, self.h, 'rgba', pixels)
        return data


cdef TTF_Font *_get_font(self) except *:
    cdef TTF_Font *fontobject = NULL
    cdef _TTFContainer ttfc
    cdef char *error
    cdef str s_error
    cdef bytes bytes_fontname

    # fast path
    fontid = self._get_font_id()
    if fontid in sdl2_cache:
        ttfc = sdl2_cache[fontid]
        return ttfc.font

    # ensure ttf is init.
    if not TTF_WasInit():
        TTF_Init()

    # try first the file if it's a filename
    fontname = self.options['font_name_r']
    bytes_fontname = <bytes>fontname.encode('utf-8')
    ext = fontname.rsplit('.', 1)
    if len(ext) == 2:
        # try to open the font if it has an extension
        fontobject = TTF_OpenFont(bytes_fontname,
                                  int(self.options['font_size']))
    # fallback to search a system font
    if fontobject == NULL:
        s_error = SDL_GetError()
        raise ValueError('{}: for font {}'.format(s_error, fontname))

    # set underline and strikethrough style
    style = TTF_STYLE_NORMAL
    if self.options['underline']:
        style = style | TTF_STYLE_UNDERLINE
    if self.options['strikethrough']:
        style = style | TTF_STYLE_STRIKETHROUGH
    TTF_SetFontStyle(fontobject, style)

    sdl2_cache[fontid] = ttfc = _TTFContainer()
    ttfc.font = fontobject
    sdl2_cache_order.append(fontid)

    # to prevent too much file open, limit the number of opened fonts to 64

    while len(sdl2_cache_order) > 64:
        popid = sdl2_cache_order.pop(0)
        ttfc = sdl2_cache[popid]
        del sdl2_cache[popid]

    ttfc = sdl2_cache[fontid]

    return ttfc.font

def _get_extents(container, text):
    cdef TTF_Font *font = _get_font(container)
    cdef int w, h
    outline_width = container.options['outline_width']
    if font == NULL:
        return 0, 0
    if not PY2:
        text = text.encode('utf-8')
    bytes_text = <bytes>text
    if outline_width:
        TTF_SetFontOutline(font, outline_width)
    TTF_SizeUTF8(font, <char *>bytes_text, &w, &h)
    if outline_width:
        TTF_SetFontOutline(font, 0)
    return w, h

def _get_fontdescent(container):
    return TTF_FontDescent(_get_font(container))

def _get_fontascent(container):
    return TTF_FontAscent(_get_font(container))



"""
CGL/GL: GL backend implementation using GL directly
"""

include "../common.pxi"
include "../../include/config.pxi"

from kivy.graphics.cgl cimport *
from kivy.logger import Logger

cdef extern from "gl_redirect.h":
    const GLubyte* (__stdcall *glGetString)(GLenum) nogil
    GLboolean (__stdcall *glIsBuffer)(GLuint buffer) nogil
    GLboolean (__stdcall *glIsEnabled)(GLenum cap) nogil
    GLboolean (__stdcall *glIsFramebuffer)(GLuint framebuffer) nogil
    GLboolean (__stdcall *glIsProgram)(GLuint program) nogil
    GLboolean (__stdcall *glIsRenderbuffer)(GLuint renderbuffer) nogil
    GLboolean (__stdcall *glIsShader)(GLuint shader) nogil
    GLboolean (__stdcall *glIsTexture)(GLuint texture) nogil
    GLenum (__stdcall *glCheckFramebufferStatus)(GLenum) nogil
    GLenum (__stdcall *glGetError)() nogil
    GLint (__stdcall *glGetAttribLocation)(GLuint, const GLchar *) nogil
    GLint (__stdcall *glGetUniformLocation)(GLuint, const char *) nogil
    GLuint (__stdcall *glCreateProgram)() nogil
    GLuint (__stdcall *glCreateShader)(GLenum) nogil
    void (__stdcall *glActiveTexture)(GLenum) nogil
    void (__stdcall *glAttachShader)(GLuint, GLuint) nogil
    void (__stdcall *glBindAttribLocation)(GLuint, GLuint, const char *) nogil
    void (__stdcall *glBindBuffer)(GLenum, GLuint) nogil
    void (__stdcall *glBindFramebuffer)(GLenum, GLuint) nogil
    void (__stdcall *glBindRenderbuffer)(GLenum target, GLuint renderbuffer) nogil
    void (__stdcall *glBindTexture)(GLenum, GLuint) nogil
    void (__stdcall *glBlendColor)(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha) nogil
    void (__stdcall *glBlendEquation)( GLenum mode ) nogil
    void (__stdcall *glBlendEquationSeparate)(GLenum modeRGB, GLenum modeAlpha) nogil
    void (__stdcall *glBlendFunc)(GLenum sfactor, GLenum dfactor) nogil
    void (__stdcall *glBlendFuncSeparate)(GLenum, GLenum, GLenum, GLenum) nogil
    void (__stdcall *glBufferData)(GLenum, GLsizeiptr, const GLvoid *, GLenum) nogil
    void (__stdcall *glBufferSubData)(GLenum, GLintptr, GLsizeiptr, const GLvoid *) nogil
    void (__stdcall *glClear)(GLbitfield) nogil
    void (__stdcall *glClearColor)(GLclampf, GLclampf, GLclampf, GLclampf) nogil
    void (__stdcall *glClearStencil)(GLint s) nogil
    void (__stdcall *glColorMask)(GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha) nogil
    void (__stdcall *glCompileShader)(GLuint) nogil
    void (__stdcall *glCompressedTexImage2D)(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLint border, GLsizei imageSize, const GLvoid* data) nogil
    void (__stdcall *glCompressedTexSubImage2D)(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLsizei imageSize, const GLvoid* data) nogil
    void (__stdcall *glCopyTexImage2D)(GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border) nogil
    void (__stdcall *glCopyTexSubImage2D)(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height) nogil
    void (__stdcall *glCullFace)(GLenum mode) nogil
    void (__stdcall *glDeleteBuffers)(GLsizei n, const GLuint* buffers) nogil
    void (__stdcall *glDeleteFramebuffers)(GLsizei, const GLuint *) nogil
    void (__stdcall *glDeleteProgram)(GLuint) nogil
    void (__stdcall *glDeleteRenderbuffers)(GLsizei n, const GLuint* renderbuffers) nogil
    void (__stdcall *glDeleteShader)(GLuint) nogil
    void (__stdcall *glDeleteTextures)(GLsizei, const GLuint *) nogil
    void (__stdcall *glDepthFunc)(GLenum func) nogil
    void (__stdcall *glDepthMask)(GLboolean flag) nogil
    void (__stdcall *glDetachShader)(GLuint program, GLuint shader) nogil
    void (__stdcall *glDisable)(GLenum) nogil
    void (__stdcall *glDisableVertexAttribArray)(GLuint) nogil
    void (__stdcall *glDrawArrays)(GLenum, GLint, GLsizei) nogil
    void (__stdcall *glDrawElements)(GLenum mode, GLsizei count, GLenum type, const GLvoid* indices) nogil
    void (__stdcall *glEnable)(GLenum) nogil
    void (__stdcall *glEnableVertexAttribArray)(GLuint) nogil
    void (__stdcall *glFinish)() nogil
    void (__stdcall *glFlush)() nogil
    void (__stdcall *glFramebufferRenderbuffer)(GLenum target, GLenum attachment, GLenum renderbuffertarget, GLuint renderbuffer) nogil
    void (__stdcall *glFramebufferTexture2D)(GLenum, GLenum, GLenum, GLuint, GLint) nogil
    void (__stdcall *glFrontFace)(GLenum mode) nogil
    void (__stdcall *glGenBuffers)(GLsizei, GLuint *) nogil
    void (__stdcall *glGenerateMipmap)(GLenum target) nogil
    void (__stdcall *glGenFramebuffers)(GLsizei, GLuint *) nogil
    void (__stdcall *glGenRenderbuffers)(GLsizei n, GLuint* renderbuffers) nogil
    void (__stdcall *glGenTextures)(GLsizei, GLuint *) nogil
    void (__stdcall *glGetActiveAttrib)(GLuint program, GLuint index, GLsizei bufsize, GLsizei* length, GLint* size, GLenum* type, GLchar* name) nogil
    void (__stdcall *glGetActiveUniform)(GLuint program, GLuint index, GLsizei bufsize, GLsizei* length, GLint* size, GLenum* type, GLchar* name) nogil
    void (__stdcall *glGetAttachedShaders)(GLuint program, GLsizei maxcount, GLsizei* count, GLuint* shaders) nogil
    void (__stdcall *glGetBooleanv)(GLenum, GLboolean *) nogil
    void (__stdcall *glGetBufferParameteriv)(GLenum target, GLenum pname, GLint* params) nogil
    void (__stdcall *glGetFloatv)(GLenum pname, GLfloat* params) nogil
    void (__stdcall *glGetFramebufferAttachmentParameteriv)(GLenum target, GLenum attachment, GLenum pname, GLint* params) nogil
    void (__stdcall *glGetIntegerv)(GLenum, GLint *) nogil
    void (__stdcall *glGetProgramInfoLog)(GLuint, GLsizei, GLsizei*, GLchar*) nogil
    void (__stdcall *glGetProgramiv)(GLuint, GLenum, GLint *) nogil
    void (__stdcall *glGetRenderbufferParameteriv)(GLenum target, GLenum pname, GLint* params) nogil
    void (__stdcall *glGetShaderInfoLog)(GLuint, GLsizei, GLsizei *, char *) nogil
    void (__stdcall *glGetShaderiv)(GLuint, GLenum, GLint *) nogil
    void (__stdcall *glGetShaderSource)(GLuint shader, GLsizei bufsize, GLsizei* length, GLchar* source) nogil
    void (__stdcall *glGetTexParameterfv)(GLenum target, GLenum pname, GLfloat* params) nogil
    void (__stdcall *glGetTexParameteriv)(GLenum target, GLenum pname, GLint* params) nogil
    void (__stdcall *glGetUniformfv)(GLuint program, GLint location, GLfloat* params) nogil
    void (__stdcall *glGetUniformiv)(GLuint program, GLint location, GLint* params) nogil
    void (__stdcall *glGetVertexAttribfv)(GLuint index, GLenum pname, GLfloat* params) nogil
    void (__stdcall *glGetVertexAttribiv)(GLuint index, GLenum pname, GLint* params) nogil
    void (__stdcall *glHint)(GLenum target, GLenum mode) nogil
    void (__stdcall *glLineWidth)(GLfloat width) nogil
    void (__stdcall *glLinkProgram)(GLuint) nogil
    void (__stdcall *glPixelStorei)(GLenum, GLint) nogil
    void (__stdcall *glPolygonOffset)(GLfloat factor, GLfloat units) nogil
    void (__stdcall *glReadPixels)(GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, GLvoid*) nogil
    void (__stdcall *glRenderbufferStorage)(GLenum target, GLenum internalformat, GLsizei width, GLsizei height) nogil
    void (__stdcall *glSampleCoverage)(GLclampf value, GLboolean invert) nogil
    void (__stdcall *glScissor)(GLint, GLint, GLsizei, GLsizei) nogil
    void (__stdcall *glShaderBinary)(GLsizei, const GLuint *, GLenum, const void *, GLsizei) nogil
    void (__stdcall *glShaderSource)(GLuint, GLsizei, const GLchar* const *, const GLint *) nogil
    void (__stdcall *glStencilFunc)(GLenum func, GLint ref, GLuint mask) nogil
    void (__stdcall *glStencilFuncSeparate)(GLenum face, GLenum func, GLint ref, GLuint mask) nogil
    void (__stdcall *glStencilMask)(GLuint mask) nogil
    void (__stdcall *glStencilMaskSeparate)(GLenum face, GLuint mask) nogil
    void (__stdcall *glStencilOp)(GLenum fail, GLenum zfail, GLenum zpass) nogil
    void (__stdcall *glStencilOpSeparate)(GLenum face, GLenum fail, GLenum zfail, GLenum zpass) nogil
    void (__stdcall *glTexImage2D)(GLenum, GLint, GLint, GLsizei, GLsizei, GLint, GLenum, GLenum, const void *) nogil
    void (__stdcall *glTexParameterf)(GLenum target, GLenum pname, GLfloat param) nogil
    void (__stdcall *glTexParameteri)(GLenum, GLenum, GLint) nogil
    void (__stdcall *glTexSubImage2D)(GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *) nogil
    void (__stdcall *glUniform1f)(GLint location, GLfloat x) nogil
    void (__stdcall *glUniform1fv)(GLint location, GLsizei count, const GLfloat* v) nogil
    void (__stdcall *glUniform1i)(GLint, GLint) nogil
    void (__stdcall *glUniform1iv)(GLint location, GLsizei count, const GLint* v) nogil
    void (__stdcall *glUniform2f)(GLint location, GLfloat x, GLfloat y) nogil
    void (__stdcall *glUniform2fv)(GLint location, GLsizei count, const GLfloat* v) nogil
    void (__stdcall *glUniform2i)(GLint location, GLint x, GLint y) nogil
    void (__stdcall *glUniform2iv)(GLint location, GLsizei count, const GLint* v) nogil
    void (__stdcall *glUniform3f)(GLint location, GLfloat x, GLfloat y, GLfloat z) nogil
    void (__stdcall *glUniform3fv)(GLint location, GLsizei count, const GLfloat* v) nogil
    void (__stdcall *glUniform3i)(GLint location, GLint x, GLint y, GLint z) nogil
    void (__stdcall *glUniform3iv)(GLint location, GLsizei count, const GLint* v) nogil
    void (__stdcall *glUniform4f)(GLint, GLfloat, GLfloat, GLfloat, GLfloat) nogil
    void (__stdcall *glUniform4fv)(GLint location, GLsizei count, const GLfloat* v) nogil
    void (__stdcall *glUniform4i)(GLint location, GLint x, GLint y, GLint z, GLint w) nogil
    void (__stdcall *glUniform4iv)(GLint location, GLsizei count, const GLint* v) nogil
    void (__stdcall *glUniformMatrix4fv)(GLint, GLsizei, GLboolean, const GLfloat *) nogil
    void (__stdcall *glUseProgram)(GLuint) nogil
    void (__stdcall *glValidateProgram)(GLuint program) nogil
    void (__stdcall *glVertexAttrib1f)(GLuint indx, GLfloat x) nogil
    void (__stdcall *glVertexAttrib2f)(GLuint indx, GLfloat x, GLfloat y) nogil
    void (__stdcall *glVertexAttrib3f)(GLuint indx, GLfloat x, GLfloat y, GLfloat z) nogil
    void (__stdcall *glVertexAttrib4f)(GLuint indx, GLfloat x, GLfloat y, GLfloat z, GLfloat w) nogil
    void (__stdcall *glVertexAttribPointer)(GLuint, GLint, GLenum, GLboolean, GLsizei, const void *) nogil
    void (__stdcall *glViewport)(GLint, GLint, GLsizei, GLsizei) nogil


cpdef is_backend_supported():
    return not USE_OPENGL_MOCK and PLATFORM != "win32"


def init_backend():
    IF USE_OPENGL_MOCK or PLATFORM == "win32":
        raise TypeError('GL is not available. Recompile with USE_OPENGL_MOCK=0')
    ELSE:
        link_static()


cpdef link_static():
    IF USE_OPENGL_MOCK:
        pass
    ELSE:
        cgl.glActiveTexture = glActiveTexture
        cgl.glAttachShader = glAttachShader
        cgl.glBindAttribLocation = glBindAttribLocation
        cgl.glBindBuffer = glBindBuffer
        cgl.glBindFramebuffer = glBindFramebuffer
        cgl.glBindRenderbuffer = glBindRenderbuffer
        cgl.glBindTexture = glBindTexture
        cgl.glBlendColor = glBlendColor
        cgl.glBlendEquation = glBlendEquation
        cgl.glBlendEquationSeparate = glBlendEquationSeparate
        cgl.glBlendFunc = glBlendFunc
        cgl.glBlendFuncSeparate = glBlendFuncSeparate
        cgl.glBufferData = glBufferData
        cgl.glBufferSubData = glBufferSubData
        cgl.glCheckFramebufferStatus = glCheckFramebufferStatus
        cgl.glClear = glClear
        cgl.glClearColor = glClearColor
        cgl.glClearStencil = glClearStencil
        cgl.glColorMask = glColorMask
        cgl.glCompileShader = glCompileShader
        cgl.glCompressedTexImage2D = glCompressedTexImage2D
        cgl.glCompressedTexSubImage2D = glCompressedTexSubImage2D
        cgl.glCopyTexImage2D = glCopyTexImage2D
        cgl.glCopyTexSubImage2D = glCopyTexSubImage2D
        cgl.glCreateProgram = glCreateProgram
        cgl.glCreateShader = glCreateShader
        cgl.glCullFace = glCullFace
        cgl.glDeleteBuffers = glDeleteBuffers
        cgl.glDeleteFramebuffers = glDeleteFramebuffers
        cgl.glDeleteProgram = glDeleteProgram
        cgl.glDeleteRenderbuffers = glDeleteRenderbuffers
        cgl.glDeleteShader = glDeleteShader
        cgl.glDeleteTextures = glDeleteTextures
        cgl.glDepthFunc = glDepthFunc
        cgl.glDepthMask = glDepthMask
        cgl.glDetachShader = glDetachShader
        cgl.glDisable = glDisable
        cgl.glDisableVertexAttribArray = glDisableVertexAttribArray
        cgl.glDrawArrays = glDrawArrays
        cgl.glDrawElements = glDrawElements
        cgl.glEnable = glEnable
        cgl.glEnableVertexAttribArray = glEnableVertexAttribArray
        cgl.glFinish = glFinish
        cgl.glFlush = glFlush
        cgl.glFramebufferRenderbuffer = glFramebufferRenderbuffer
        cgl.glFramebufferTexture2D = glFramebufferTexture2D
        cgl.glFrontFace = glFrontFace
        cgl.glGenBuffers = glGenBuffers
        cgl.glGenerateMipmap = glGenerateMipmap
        cgl.glGenFramebuffers = glGenFramebuffers
        cgl.glGenRenderbuffers = glGenRenderbuffers
        cgl.glGenTextures = glGenTextures
        cgl.glGetActiveAttrib = glGetActiveAttrib
        cgl.glGetActiveUniform = glGetActiveUniform
        cgl.glGetAttachedShaders = glGetAttachedShaders
        cgl.glGetAttribLocation = glGetAttribLocation
        cgl.glGetBooleanv = glGetBooleanv
        cgl.glGetBufferParameteriv = glGetBufferParameteriv
        cgl.glGetError = glGetError
        cgl.glGetFloatv = glGetFloatv
        cgl.glGetFramebufferAttachmentParameteriv = glGetFramebufferAttachmentParameteriv
        cgl.glGetIntegerv = glGetIntegerv
        cgl.glGetProgramInfoLog = glGetProgramInfoLog
        cgl.glGetProgramiv = glGetProgramiv
        cgl.glGetRenderbufferParameteriv = glGetRenderbufferParameteriv
        cgl.glGetShaderInfoLog = glGetShaderInfoLog
        cgl.glGetShaderiv = glGetShaderiv
        cgl.glGetShaderSource = glGetShaderSource
        cgl.glGetString = glGetString
        cgl.glGetTexParameterfv = glGetTexParameterfv
        cgl.glGetTexParameteriv = glGetTexParameteriv
        cgl.glGetUniformfv = glGetUniformfv
        cgl.glGetUniformiv = glGetUniformiv
        cgl.glGetUniformLocation = glGetUniformLocation
        cgl.glGetVertexAttribfv = glGetVertexAttribfv
        cgl.glGetVertexAttribiv = glGetVertexAttribiv
        cgl.glHint = glHint
        cgl.glIsBuffer = glIsBuffer
        cgl.glIsEnabled = glIsEnabled
        cgl.glIsFramebuffer = glIsFramebuffer
        cgl.glIsProgram = glIsProgram
        cgl.glIsRenderbuffer = glIsRenderbuffer
        cgl.glIsShader = glIsShader
        cgl.glIsTexture = glIsTexture
        cgl.glLineWidth = glLineWidth
        cgl.glLinkProgram = glLinkProgram
        cgl.glPixelStorei = glPixelStorei
        cgl.glPolygonOffset = glPolygonOffset
        cgl.glReadPixels = glReadPixels
        cgl.glRenderbufferStorage = glRenderbufferStorage
        cgl.glSampleCoverage = glSampleCoverage
        cgl.glScissor = glScissor
        # cgl.glShaderBinary = glShaderBinary
        cgl.glShaderSource = glShaderSource
        cgl.glStencilFunc = glStencilFunc
        cgl.glStencilFuncSeparate = glStencilFuncSeparate
        cgl.glStencilMask = glStencilMask
        cgl.glStencilMaskSeparate = glStencilMaskSeparate
        cgl.glStencilOp = glStencilOp
        cgl.glStencilOpSeparate = glStencilOpSeparate
        cgl.glTexImage2D = glTexImage2D
        cgl.glTexParameterf = glTexParameterf
        cgl.glTexParameteri = glTexParameteri
        cgl.glTexSubImage2D = glTexSubImage2D
        cgl.glUniform1f = glUniform1f
        cgl.glUniform1fv = glUniform1fv
        cgl.glUniform1i = glUniform1i
        cgl.glUniform1iv = glUniform1iv
        cgl.glUniform2f = glUniform2f
        cgl.glUniform2fv = glUniform2fv
        cgl.glUniform2i = glUniform2i
        cgl.glUniform2iv = glUniform2iv
        cgl.glUniform3f = glUniform3f
        cgl.glUniform3fv = glUniform3fv
        cgl.glUniform3i = glUniform3i
        cgl.glUniform3iv = glUniform3iv
        cgl.glUniform4f = glUniform4f
        cgl.glUniform4fv = glUniform4fv
        cgl.glUniform4i = glUniform4i
        cgl.glUniform4iv = glUniform4iv
        cgl.glUniformMatrix4fv = glUniformMatrix4fv
        cgl.glUseProgram = glUseProgram
        cgl.glValidateProgram = glValidateProgram
        cgl.glVertexAttrib1f = glVertexAttrib1f
        cgl.glVertexAttrib2f = glVertexAttrib2f
        cgl.glVertexAttrib3f = glVertexAttrib3f
        cgl.glVertexAttrib4f = glVertexAttrib4f
        cgl.glVertexAttribPointer = glVertexAttribPointer
        cgl.glViewport = glViewport



"""
CGL/Mock: GL backend implementation by mocking functions to NOOP
"""

include "../common.pxi"

from kivy.graphics.cgl cimport *

cdef GLubyte *empty_str = ''

cpdef is_backend_supported():
    return True

cdef GLenum __stdcall mockCheckFramebufferStatus(GLenum target) nogil:
    return GL_FRAMEBUFFER_COMPLETE
cdef GLuint __stdcall mockCreateProgram() nogil:
    return <GLuint>1
cdef GLuint __stdcall mockCreateShader(GLenum type) nogil:
    return <GLuint>1
cdef int __stdcall mockGetAttribLocation(GLuint program, const GLchar* name) nogil:
    return 1
cdef GLenum __stdcall mockGetError() nogil:
    return GL_NO_ERROR
cdef const GLubyte* __stdcall mockGetString(GLenum name) nogil:
    return empty_str
cdef int __stdcall mockGetUniformLocation(GLuint program,  const GLchar* name) nogil:
    1
cdef GLboolean __stdcall mockIsBuffer(GLuint buffer) nogil:
    return GL_TRUE
cdef GLboolean __stdcall  mockIsEnabled(GLenum cap) nogil:
    return GL_TRUE
cdef GLboolean __stdcall mockIsFramebuffer(GLuint framebuffer) nogil:
    return GL_TRUE
cdef GLboolean __stdcall mockIsProgram(GLuint program) nogil:
    return GL_TRUE
cdef GLboolean __stdcall mockIsRenderbuffer(GLuint renderbuffer) nogil:
    return GL_TRUE
cdef GLboolean __stdcall mockIsShader(GLuint shader) nogil:
    return GL_TRUE
cdef GLboolean __stdcall mockIsTexture(GLuint texture) nogil:
    return GL_TRUE

cdef void __stdcall mockActiveTexture(GLenum texture) nogil:
    pass
cdef void __stdcall mockAttachShader(GLuint program, GLuint shader) nogil:
    pass
cdef void __stdcall mockBindAttribLocation(GLuint program, GLuint index, const GLchar* name) nogil:
    pass
cdef void __stdcall mockBindBuffer(GLenum target, GLuint buffer) nogil:
    pass
cdef void __stdcall mockBindFramebuffer(GLenum target, GLuint framebuffer) nogil:
    pass
cdef void __stdcall mockBindRenderbuffer(GLenum target, GLuint renderbuffer) nogil:
    pass
cdef void __stdcall mockBindTexture(GLenum target, GLuint texture) nogil:
    pass
cdef void __stdcall mockBlendColor(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha) nogil:
    pass
cdef void __stdcall mockBlendEquation( GLenum mode ) nogil:
    pass
cdef void __stdcall mockBlendEquationSeparate(GLenum modeRGB, GLenum modeAlpha) nogil:
    pass
cdef void __stdcall mockBlendFunc(GLenum sfactor, GLenum dfactor) nogil:
    pass
cdef void __stdcall mockBlendFuncSeparate(GLenum srcRGB, GLenum dstRGB, GLenum srcAlpha, GLenum dstAlpha) nogil:
    pass
cdef void __stdcall mockBufferData(GLenum target, GLsizeiptr size, const GLvoid* data, GLenum usage) nogil:
    pass
cdef void __stdcall mockBufferSubData(GLenum target, GLintptr offset, GLsizeiptr size, const GLvoid* data) nogil:
    pass
cdef void __stdcall mockClear(GLbitfield mask) nogil:
    pass
cdef void __stdcall mockClearColor(GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha) nogil:
    pass
# cdef void __stdcall mockClearDepthf(GLclampf depth) nogil:
#     pass
cdef void __stdcall mockClearStencil(GLint s) nogil:
    pass
cdef void __stdcall mockColorMask(GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha) nogil:
    pass
cdef void __stdcall mockCompileShader(GLuint shader) nogil:
    pass
cdef void __stdcall mockCompressedTexImage2D(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLint border, GLsizei imageSize, const GLvoid* data) nogil:
    pass
cdef void __stdcall mockCompressedTexSubImage2D(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLsizei imageSize, const GLvoid* data) nogil:
    pass
cdef void __stdcall mockCopyTexImage2D(GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border) nogil:
    pass
cdef void __stdcall mockCopyTexSubImage2D(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height) nogil:
    pass
cdef void __stdcall mockCullFace(GLenum mode) nogil:
    pass
cdef void __stdcall mockDeleteBuffers(GLsizei n, const GLuint* buffers) nogil:
    pass
cdef void __stdcall mockDeleteFramebuffers(GLsizei n, const GLuint* framebuffers) nogil:
    pass
cdef void __stdcall mockDeleteProgram(GLuint program) nogil:
    pass
cdef void __stdcall mockDeleteRenderbuffers(GLsizei n, const GLuint* renderbuffers) nogil:
    pass
cdef void __stdcall mockDeleteShader(GLuint shader) nogil:
    pass
cdef void __stdcall mockDeleteTextures(GLsizei n, const GLuint* textures) nogil:
    pass
cdef void __stdcall mockDepthFunc(GLenum func) nogil:
    pass
cdef void __stdcall mockDepthMask(GLboolean flag) nogil:
    pass
# cdef void __stdcall mockDepthRangef(GLclampf zNear, GLclampf zFar) nogil:
#     pass
cdef void __stdcall mockDetachShader(GLuint program, GLuint shader) nogil:
    pass
cdef void __stdcall mockDisable(GLenum cap) nogil:
    pass
cdef void __stdcall mockDisableVertexAttribArray(GLuint index) nogil:
    pass
cdef void __stdcall mockDrawArrays(GLenum mode, GLint first, GLsizei count) nogil:
    pass
cdef void __stdcall mockDrawElements(GLenum mode, GLsizei count, GLenum type, const GLvoid* indices) nogil:
    pass
cdef void __stdcall mockEnable(GLenum cap) nogil:
    pass
cdef void __stdcall mockEnableVertexAttribArray(GLuint index) nogil:
    pass
cdef void __stdcall mockFinish() nogil:
    pass
cdef void __stdcall mockFlush() nogil:
    pass
cdef void __stdcall mockFramebufferRenderbuffer(GLenum target, GLenum attachment, GLenum renderbuffertarget, GLuint renderbuffer) nogil:
    pass
cdef void __stdcall mockFramebufferTexture2D(GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level) nogil:
    pass
cdef void __stdcall mockFrontFace(GLenum mode) nogil:
    pass
cdef void __stdcall mockGenBuffers(GLsizei n, GLuint* buffers) nogil:
    pass
cdef void __stdcall mockGenerateMipmap(GLenum target) nogil:
    pass
cdef void __stdcall mockGenFramebuffers(GLsizei n, GLuint* framebuffers) nogil:
    pass
cdef void __stdcall mockGenRenderbuffers(GLsizei n, GLuint* renderbuffers) nogil:
    pass
cdef void __stdcall mockGenTextures(GLsizei n, GLuint* textures) nogil:
    pass
cdef void __stdcall mockGetActiveAttrib(GLuint program, GLuint index, GLsizei bufsize, GLsizei* length, GLint* size, GLenum* type, GLchar* name) nogil:
    pass
cdef void __stdcall mockGetActiveUniform(GLuint program, GLuint index, GLsizei bufsize, GLsizei* length, GLint* size, GLenum* type, GLchar* name) nogil:
    pass
cdef void __stdcall mockGetAttachedShaders(GLuint program, GLsizei maxcount, GLsizei* count, GLuint* shaders) nogil:
    pass
cdef void __stdcall mockGetBooleanv(GLenum pname, GLboolean* params) nogil:
    pass
cdef void __stdcall mockGetBufferParameteriv(GLenum target, GLenum pname, GLint* params) nogil:
    pass
cdef void __stdcall mockGetFloatv(GLenum pname, GLfloat* params) nogil:
    pass
cdef void __stdcall mockGetFramebufferAttachmentParameteriv(GLenum target, GLenum attachment, GLenum pname, GLint* params) nogil:
    pass
cdef void __stdcall mockGetIntegerv(GLenum pname, GLint* params) nogil:
    pass
cdef void __stdcall mockGetProgramiv(GLuint program, GLenum pname, GLint* params) nogil:
    pass
cdef void __stdcall mockGetProgramInfoLog(GLuint program, GLsizei bufsize, GLsizei* length, GLchar* infolog) nogil:
    pass
cdef void __stdcall mockGetRenderbufferParameteriv(GLenum target, GLenum pname, GLint* params) nogil:
    pass
cdef void __stdcall mockGetShaderiv(GLuint shader, GLenum pname, GLint* params) nogil:
    pass
cdef void __stdcall mockGetShaderInfoLog(GLuint shader, GLsizei bufsize, GLsizei* length, GLchar* infolog) nogil:
    pass
# cdef void __stdcall mockGetShaderPrecisionFormat(GLenum shadertype, GLenum precisiontype, GLint* range, GLint* precision) nogil:
#     pass
cdef void __stdcall mockGetShaderSource(GLuint shader, GLsizei bufsize, GLsizei* length, GLchar* source) nogil:
    pass
cdef void __stdcall mockGetTexParameterfv(GLenum target, GLenum pname, GLfloat* params) nogil:
    pass
cdef void __stdcall mockGetTexParameteriv(GLenum target, GLenum pname, GLint* params) nogil:
    pass
cdef void __stdcall mockGetUniformfv(GLuint program, GLint location, GLfloat* params) nogil:
    pass
cdef void __stdcall mockGetUniformiv(GLuint program, GLint location, GLint* params) nogil:
    pass
cdef void __stdcall mockGetVertexAttribfv(GLuint index, GLenum pname, GLfloat* params) nogil:
    pass
cdef void __stdcall mockGetVertexAttribiv(GLuint index, GLenum pname, GLint* params) nogil:
    pass
# cdef void __stdcall mockGetVertexAttribPointerv(GLuint index, GLenum pname, GLvoid** pointer) nogil:
#     pass
cdef void __stdcall mockHint(GLenum target, GLenum mode) nogil:
    pass
cdef void __stdcall mockLineWidth(GLfloat width) nogil:
    pass
cdef void __stdcall mockLinkProgram(GLuint program) nogil:
    pass
cdef void __stdcall mockPixelStorei(GLenum pname, GLint param) nogil:
    pass
cdef void __stdcall mockPolygonOffset(GLfloat factor, GLfloat units) nogil:
    pass
cdef void __stdcall mockReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid* pixels) nogil:
    pass
# cdef void __stdcall mockReleaseShaderCompiler() nogil:
#     pass
cdef void __stdcall mockRenderbufferStorage(GLenum target, GLenum internalformat, GLsizei width, GLsizei height) nogil:
    pass
cdef void __stdcall mockSampleCoverage(GLclampf value, GLboolean invert) nogil:
    pass
cdef void __stdcall mockScissor(GLint x, GLint y, GLsizei width, GLsizei height) nogil:
    pass
cdef void __stdcall mockShaderBinary(GLsizei n, const GLuint* shaders, GLenum binaryformat, const GLvoid* binary, GLsizei length) nogil:
    pass
cdef void __stdcall mockShaderSource(GLuint shader, GLsizei count, const GLchar* const* string, const GLint* length) nogil:
    pass
cdef void __stdcall mockStencilFunc(GLenum func, GLint ref, GLuint mask) nogil:
    pass
cdef void __stdcall mockStencilFuncSeparate(GLenum face, GLenum func, GLint ref, GLuint mask) nogil:
    pass
cdef void __stdcall mockStencilMask(GLuint mask) nogil:
    pass
cdef void __stdcall mockStencilMaskSeparate(GLenum face, GLuint mask) nogil:
    pass
cdef void __stdcall mockStencilOp(GLenum fail, GLenum zfail, GLenum zpass) nogil:
    pass
cdef void __stdcall mockStencilOpSeparate(GLenum face, GLenum fail, GLenum zfail, GLenum zpass) nogil:
    pass
cdef void __stdcall mockTexImage2D(GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid* pixels) nogil:
    pass
cdef void __stdcall mockTexParameterf(GLenum target, GLenum pname, GLfloat param) nogil:
    pass
# cdef void __stdcall mockTexParameterfv(GLenum target, GLenum pname, GLfloat* params) nogil:
#     pass
cdef void __stdcall mockTexParameteri(GLenum target, GLenum pname, GLint param) nogil:
    pass
# cdef void __stdcall mockTexParameteriv(GLenum target, GLenum pname, GLint* params) nogil:
#     pass
cdef void __stdcall mockTexSubImage2D(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid* pixels) nogil:
    pass
cdef void __stdcall mockUniform1f(GLint location, GLfloat x) nogil:
    pass
cdef void __stdcall mockUniform1fv(GLint location, GLsizei count, const GLfloat* v) nogil:
    pass
cdef void __stdcall mockUniform1i(GLint location, GLint x) nogil:
    pass
cdef void __stdcall mockUniform1iv(GLint location, GLsizei count, const GLint* v) nogil:
    pass
cdef void __stdcall mockUniform2f(GLint location, GLfloat x, GLfloat y) nogil:
    pass
cdef void __stdcall mockUniform2fv(GLint location, GLsizei count, const GLfloat* v) nogil:
    pass
cdef void __stdcall mockUniform2i(GLint location, GLint x, GLint y) nogil:
    pass
cdef void __stdcall mockUniform2iv(GLint location, GLsizei count, const GLint* v) nogil:
    pass
cdef void __stdcall mockUniform3f(GLint location, GLfloat x, GLfloat y, GLfloat z) nogil:
    pass
cdef void __stdcall mockUniform3fv(GLint location, GLsizei count, const GLfloat* v) nogil:
    pass
cdef void __stdcall mockUniform3i(GLint location, GLint x, GLint y, GLint z) nogil:
    pass
cdef void __stdcall mockUniform3iv(GLint location, GLsizei count, const GLint* v) nogil:
    pass
cdef void __stdcall mockUniform4f(GLint location, GLfloat x, GLfloat y, GLfloat z, GLfloat w) nogil:
    pass
cdef void __stdcall mockUniform4fv(GLint location, GLsizei count, const GLfloat* v) nogil:
    pass
cdef void __stdcall mockUniform4i(GLint location, GLint x, GLint y, GLint z, GLint w) nogil:
    pass
cdef void __stdcall mockUniform4iv(GLint location, GLsizei count, const GLint* v) nogil:
    pass
# cdef void __stdcall mockUniformMatrix2fv(GLint location, GLsizei count, GLboolean transpose, const GLfloat* value) nogil:
#     pass
# cdef void __stdcall mockUniformMatrix3fv(GLint location, GLsizei count, GLboolean transpose, const GLfloat* value) nogil:
#     pass
cdef void __stdcall mockUniformMatrix4fv(GLint location, GLsizei count, GLboolean transpose, const GLfloat* value) nogil:
    pass
cdef void __stdcall mockUseProgram(GLuint program) nogil:
    pass
cdef void __stdcall mockValidateProgram(GLuint program) nogil:
    pass
cdef void __stdcall mockVertexAttrib1f(GLuint indx, GLfloat x) nogil:
    pass
# cdef void __stdcall mockVertexAttrib1fv(GLuint indx, GLfloat* values) nogil:
#     pass
cdef void __stdcall mockVertexAttrib2f(GLuint indx, GLfloat x, GLfloat y) nogil:
    pass
# cdef void __stdcall mockVertexAttrib2fv(GLuint indx, GLfloat* values) nogil:
#     pass
cdef void __stdcall mockVertexAttrib3f(GLuint indx, GLfloat x, GLfloat y, GLfloat z) nogil:
    pass
# cdef void __stdcall mockVertexAttrib3fv(GLuint indx, GLfloat* values) nogil:
#     pass
cdef void __stdcall mockVertexAttrib4f(GLuint indx, GLfloat x, GLfloat y, GLfloat z, GLfloat w) nogil:
    pass
# cdef void __stdcall mockVertexAttrib4fv(GLuint indx, GLfloat* values) nogil:
#     pass
cdef void __stdcall mockVertexAttribPointer(GLuint indx, GLint size, GLenum type, GLboolean normalized, GLsizei stride, const GLvoid* ptr) nogil:
    pass
cdef void __stdcall mockViewport(GLint x, GLint y, GLsizei width, GLsizei height) nogil:
    pass


def init_backend():
    cgl.glActiveTexture = mockActiveTexture
    cgl.glAttachShader = mockAttachShader
    cgl.glBindAttribLocation = mockBindAttribLocation
    cgl.glBindBuffer = mockBindBuffer
    cgl.glBindFramebuffer = mockBindFramebuffer
    cgl.glBindRenderbuffer = mockBindRenderbuffer
    cgl.glBindTexture = mockBindTexture
    cgl.glBlendColor = mockBlendColor
    cgl.glBlendEquation = mockBlendEquation
    cgl.glBlendEquationSeparate = mockBlendEquationSeparate
    cgl.glBlendFunc = mockBlendFunc
    cgl.glBlendFuncSeparate = mockBlendFuncSeparate
    cgl.glBufferData = mockBufferData
    cgl.glBufferSubData = mockBufferSubData
    cgl.glCheckFramebufferStatus = mockCheckFramebufferStatus
    cgl.glClear = mockClear
    cgl.glClearColor = mockClearColor
    cgl.glClearStencil = mockClearStencil
    cgl.glColorMask = mockColorMask
    cgl.glCompileShader = mockCompileShader
    cgl.glCompressedTexImage2D = mockCompressedTexImage2D
    cgl.glCompressedTexSubImage2D = mockCompressedTexSubImage2D
    cgl.glCopyTexImage2D = mockCopyTexImage2D
    cgl.glCopyTexSubImage2D = mockCopyTexSubImage2D
    cgl.glCreateProgram = mockCreateProgram
    cgl.glCreateShader = mockCreateShader
    cgl.glCullFace = mockCullFace
    cgl.glDeleteBuffers = mockDeleteBuffers
    cgl.glDeleteFramebuffers = mockDeleteFramebuffers
    cgl.glDeleteProgram = mockDeleteProgram
    cgl.glDeleteRenderbuffers = mockDeleteRenderbuffers
    cgl.glDeleteShader = mockDeleteShader
    cgl.glDeleteTextures = mockDeleteTextures
    cgl.glDepthFunc = mockDepthFunc
    cgl.glDepthMask = mockDepthMask
    cgl.glDetachShader = mockDetachShader
    cgl.glDisable = mockDisable
    cgl.glDisableVertexAttribArray = mockDisableVertexAttribArray
    cgl.glDrawArrays = mockDrawArrays
    cgl.glDrawElements = mockDrawElements
    cgl.glEnable = mockEnable
    cgl.glEnableVertexAttribArray = mockEnableVertexAttribArray
    cgl.glFinish = mockFinish
    cgl.glFlush = mockFlush
    cgl.glFramebufferRenderbuffer = mockFramebufferRenderbuffer
    cgl.glFramebufferTexture2D = mockFramebufferTexture2D
    cgl.glFrontFace = mockFrontFace
    cgl.glGenBuffers = mockGenBuffers
    cgl.glGenerateMipmap = mockGenerateMipmap
    cgl.glGenFramebuffers = mockGenFramebuffers
    cgl.glGenRenderbuffers = mockGenRenderbuffers
    cgl.glGenTextures = mockGenTextures
    cgl.glGetActiveAttrib = mockGetActiveAttrib
    cgl.glGetActiveUniform = mockGetActiveUniform
    cgl.glGetAttachedShaders = mockGetAttachedShaders
    cgl.glGetAttribLocation = mockGetAttribLocation
    cgl.glGetBooleanv = mockGetBooleanv
    cgl.glGetBufferParameteriv = mockGetBufferParameteriv
    cgl.glGetError = mockGetError
    cgl.glGetFloatv = mockGetFloatv
    cgl.glGetFramebufferAttachmentParameteriv = mockGetFramebufferAttachmentParameteriv
    cgl.glGetIntegerv = mockGetIntegerv
    cgl.glGetProgramInfoLog = mockGetProgramInfoLog
    cgl.glGetProgramiv = mockGetProgramiv
    cgl.glGetRenderbufferParameteriv = mockGetRenderbufferParameteriv
    cgl.glGetShaderInfoLog = mockGetShaderInfoLog
    cgl.glGetShaderiv = mockGetShaderiv
    cgl.glGetShaderSource = mockGetShaderSource
    cgl.glGetString = mockGetString
    cgl.glGetTexParameterfv = mockGetTexParameterfv
    cgl.glGetTexParameteriv = mockGetTexParameteriv
    cgl.glGetUniformfv = mockGetUniformfv
    cgl.glGetUniformiv = mockGetUniformiv
    cgl.glGetUniformLocation = mockGetUniformLocation
    cgl.glGetVertexAttribfv = mockGetVertexAttribfv
    cgl.glGetVertexAttribiv = mockGetVertexAttribiv
    cgl.glHint = mockHint
    cgl.glIsBuffer = mockIsBuffer
    cgl.glIsEnabled = mockIsEnabled
    cgl.glIsFramebuffer = mockIsFramebuffer
    cgl.glIsProgram = mockIsProgram
    cgl.glIsRenderbuffer = mockIsRenderbuffer
    cgl.glIsShader = mockIsShader
    cgl.glIsTexture = mockIsTexture
    cgl.glLineWidth = mockLineWidth
    cgl.glLinkProgram = mockLinkProgram
    cgl.glPixelStorei = mockPixelStorei
    cgl.glPolygonOffset = mockPolygonOffset
    cgl.glReadPixels = mockReadPixels
    cgl.glRenderbufferStorage = mockRenderbufferStorage
    cgl.glSampleCoverage = mockSampleCoverage
    cgl.glScissor = mockScissor
    cgl.glShaderBinary = mockShaderBinary
    cgl.glShaderSource = mockShaderSource
    cgl.glStencilFunc = mockStencilFunc
    cgl.glStencilFuncSeparate = mockStencilFuncSeparate
    cgl.glStencilMask = mockStencilMask
    cgl.glStencilMaskSeparate = mockStencilMaskSeparate
    cgl.glStencilOp = mockStencilOp
    cgl.glStencilOpSeparate = mockStencilOpSeparate
    cgl.glTexImage2D = mockTexImage2D
    cgl.glTexParameterf = mockTexParameterf
    cgl.glTexParameteri = mockTexParameteri
    cgl.glTexSubImage2D = mockTexSubImage2D
    cgl.glUniform1f = mockUniform1f
    cgl.glUniform1fv = mockUniform1fv
    cgl.glUniform1i = mockUniform1i
    cgl.glUniform1iv = mockUniform1iv
    cgl.glUniform2f = mockUniform2f
    cgl.glUniform2fv = mockUniform2fv
    cgl.glUniform2i = mockUniform2i
    cgl.glUniform2iv = mockUniform2iv
    cgl.glUniform3f = mockUniform3f
    cgl.glUniform3fv = mockUniform3fv
    cgl.glUniform3i = mockUniform3i
    cgl.glUniform3iv = mockUniform3iv
    cgl.glUniform4f = mockUniform4f
    cgl.glUniform4fv = mockUniform4fv
    cgl.glUniform4i = mockUniform4i
    cgl.glUniform4iv = mockUniform4iv
    cgl.glUniformMatrix4fv = mockUniformMatrix4fv
    cgl.glUseProgram = mockUseProgram
    cgl.glValidateProgram = mockValidateProgram
    cgl.glVertexAttrib1f = mockVertexAttrib1f
    cgl.glVertexAttrib2f = mockVertexAttrib2f
    cgl.glVertexAttrib3f = mockVertexAttrib3f
    cgl.glVertexAttrib4f = mockVertexAttrib4f
    cgl.glVertexAttribPointer = mockVertexAttribPointer
    cgl.glViewport = mockViewport



'''
SVG
===

.. versionadded:: 1.9.0

.. warning::

    This is highly experimental and subject to change. Don't use it in
    production.


Load an SVG as a graphics instruction::

    from kivy.graphics.svg import Svg
    with widget.canvas:
        svg = Svg("image.svg")

There is no widget that can display Svg directly, you have to make your own for
now. Check the `examples/svg` for more information.
'''

__all__ = ("Svg", )

include "common.pxi"

import math
import re
cimport cython
from xml.etree.cElementTree import parse
from kivy.graphics.instructions cimport RenderContext
from kivy.graphics.vertex_instructions cimport Mesh, StripMesh
from kivy.graphics.tesselator cimport Tesselator
from kivy.graphics.texture cimport Texture
from kivy.graphics.vertex cimport VertexFormat
from kivy.logger import Logger
from cpython cimport array
from array import array
from cython cimport view
from time import time
from kivy.utils import hex_colormap
from kivy.properties import NUMERIC_FORMATS, dpi2px
from string import hexdigits
from kivy.core.window import Window

cdef dict colormap = hex_colormap

DEF BEZIER_POINTS = 64 # 10
DEF CIRCLE_POINTS = 64 # 24
DEF TOLERANCE = 0.001

cdef str SVG_FS = '''
#ifdef GL_ES
    precision highp float;
#endif

varying vec4 vertex_color;
varying vec2 texcoord;
uniform sampler2D texture0;

void main (void) {
    gl_FragColor = texture2D(texture0, texcoord) * (vertex_color / 255.);
}
'''

cdef str SVG_VS = '''
#ifdef GL_ES
    precision highp float;
#endif

attribute vec2 v_pos;
attribute vec2 v_tex;
attribute vec4 v_color;
uniform mat4 modelview_mat;
uniform mat4 projection_mat;
varying vec4 vertex_color;
varying vec2 texcoord;

void main (void) {
    vertex_color = v_color;
    gl_Position = projection_mat * modelview_mat * vec4(v_pos, 0.0, 1.0);
    texcoord = v_tex;
}
'''

cdef set COMMANDS = set('MmZzLlHhVvCcSsQqTtAa')
cdef set UPPERCASE = set('MZLHVCSQTA')
cdef object RE_LIST = re.compile(
    r'([A-Za-z]|-?[0-9]+\.?[0-9]*(?:e-?[0-9]*)?)')
cdef object RE_COMMAND = re.compile(
    r'([MmZzLlHhVvCcSsQqTtAa])')
cdef object RE_FLOAT = re.compile(
    r'[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?')
cdef object RE_POLYLINE = re.compile(
    r'(-?[0-9]+\.?[0-9]*(?:e-?[0-9]*)?)')
cdef object RE_TRANSFORM = re.compile(
    r'[a-zA-Z]+\([^)]*\)')

cdef VertexFormat VERTEX_FORMAT = VertexFormat(
    (b'v_pos', 2, 'float'),
    (b'v_tex', 2, 'float'),
    (b'v_color', 4, 'float'))

def _tokenize_path(pathdef):
    for x in RE_COMMAND.split(pathdef):
        if x in COMMANDS:
            yield x
        for token in RE_FLOAT.findall(x):
            yield token

cdef inline double angle(double ux, double uy, double vx, double vy):
    a = acos((ux * vx + uy * vy) / sqrt((ux ** 2 + uy ** 2) * (vx ** 2 + vy ** 2)))
    sgn = 1 if ux * vy > uy * vx else -1
    return sgn * a

cdef float parse_width(txt, float vbox_width):
    if txt.endswith('%'):
        return <float>(vbox_width * txt[:-1] / 100.)
    return parse_float(txt)

cdef float parse_height(txt, float vbox_height):
    if txt.endswith('%'):
        return <float>(vbox_height * txt[:-1] / 100.)
    return parse_float(txt)

cdef float parse_float(txt):
    if not txt:
        return 0.
    if txt[-2:] in NUMERIC_FORMATS:
        return dpi2px(txt[:-2], txt[-2:])
    return <float>float(txt)

cdef list parse_list(string):
    return re.findall(RE_LIST, string)

cdef dict parse_style(string):
    cdef dict sdict = {}
    for item in string.split(';'):
        if ':' in item:
            key, value = item.split(':', 1)
            sdict[key] = value.strip()
    return sdict

cdef list kv_color_to_int_color(color):
    c = [int(255*x) for x in color]
    return c if len(c) == 4 else c + [255]

cdef int_color_to_kv_color(color):
    c = [int(x)/255.0 for x in color]
    return c if len(c) == 4 else c + [255]

cdef parse_color(c, current_color=None):
    cdef int r, g, b, a
    if c is None or c == 'none':
        return None
    if c[0] == '#':
        c = c[1:]
    if c[:5] == 'url(#':
        return c[5:-1]
    if str(c) == 'currentColor':
        if current_color is None:
            c = 'black'
        else:
            return current_color
    if str(c) in colormap:
        c = colormap[str(c)][1:]
        r = int(c[0:2], 16)
        g = int(c[2:4], 16)
        b = int(c[4:6], 16)
        a = 255
    elif c.startswith('rgba('):
        r, g, b, a = [int(x) for x in c[4:-1].split(',')]

    elif c.startswith('rgb('):
        r, g, b = [int(x) for x in c[4:-1].split(',')]
        a = 255

    elif all(x in hexdigits for x in c):
        if len(c) == 8:
            r = int(c[0:2], 16)
            g = int(c[2:4], 16)
            b = int(c[4:6], 16)
            a = int(c[6:8], 16)
        elif len(c) == 6:
            r = int(c[0:2], 16)
            g = int(c[2:4], 16)
            b = int(c[4:6], 16)
            a = 255
        elif len(c) == 4:
            r = int(c[0], 16) * 17
            g = int(c[1], 16) * 17
            b = int(c[2], 16) * 17
            a = int(c[3], 16) * 17
        elif len(c) == 3:
            r = int(c[0], 16) * 17
            g = int(c[1], 16) * 17
            b = int(c[2], 16) * 17
            a = 255
        else:
            raise Exception('Invalid color format {}'.format(c))
    else:
        raise Exception('Unknown color {}'.format(c))
    return [r, g, b, a]

cdef class Matrix(object):
    def __cinit__(self):
        memset(self.mat, 0, sizeof(matrix_t))

    def __init__(self, string=None):
        cdef float f
        cdef int i
        self.mat[0] = self.mat[3] = 1.
        if isinstance(string, str):
            if string.startswith('matrix('):
                i = 0
                for sf in parse_list(string[7:-1]):
                    self.mat[i] = <float>float(sf)
                    i += 1
            elif string.startswith('translate('):
                a, b = parse_list(string[10:-1])
                self.mat[4] = <float>float(a)
                self.mat[5] = <float>float(b)
            elif string.startswith('scale('):
                value = parse_list(string[6:-1])
                if len(value) == 1:
                    a = b = value[0]
                elif len(value) == 2:
                    a, b = value
                else:
                    print("SVG: unknown how to parse: {!r}".format(value))
                self.mat[0] = <float>float(a)
                self.mat[3] = <float>float(b)
            elif string.startswith('rotate('):
                value = parse_list(string[7:-1])
                angle = <float>float(value[0])
                if len(value) == 3:
                    cx, cy = map(float, value[1:])
                else:
                    cx = cy = 0
                cos_a = math.cos(math.radians(angle))
                sin_a = math.sin(math.radians(angle))
                self.mat[0] = cos_a
                self.mat[1] = sin_a
                self.mat[2] = -sin_a
                self.mat[3] = cos_a
                self.mat[4] = -cx * cos_a + cy * sin_a + cx
                self.mat[5] = -cx * sin_a - cy * cos_a + cy
        elif string is not None:
            i = 0
            for f in string:
                self.mat[i] = f
                i += 1

    cdef void transform(self, float ox, float oy, float *x, float *y):
        cdef double rx = self.mat[0] * ox + self.mat[2] * oy + self.mat[4]
        cdef double ry = self.mat[1] * ox + self.mat[3] * oy + self.mat[5]
        x[0] = <float>rx
        y[0] = <float>ry

    cpdef Matrix inverse(self):
        cdef double d = self.mat[0] * self.mat[3] - self.mat[1]*self.mat[2]
        return Matrix([self.mat[3] / d, -self.mat[1] / d, -self.mat[2] / d, self.mat[0] / d,
                       (self.mat[2] * self.mat[5] - self.mat[3] * self.mat[4]) / d,
                       (self.mat[1] * self.mat[4] - self.mat[0] * self.mat[5]) / d])

    def __mul__(Matrix self, Matrix other):
        return Matrix([
            self.mat[0] * other.mat[0] + self.mat[2] * other.mat[1],
            self.mat[1] * other.mat[0] + self.mat[3] * other.mat[1],
            self.mat[0] * other.mat[2] + self.mat[2] * other.mat[3],
            self.mat[1] * other.mat[2] + self.mat[3] * other.mat[3],
            self.mat[0] * other.mat[4] + self.mat[2] * other.mat[5] + self.mat[4],
            self.mat[1] * other.mat[4] + self.mat[3] * other.mat[5] + self.mat[5]])


class GradientContainer(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.callback_dict = {}

    def call_me_on_add(self, callback, grad_id):
        '''The client wants to know when the gradient with id grad_id gets
        added.  So store this callback for when that happens.
        When the desired gradient is added, the callback will be called
        with the gradient as the first and only argument.
        '''
        cblist = self.callback_dict.get(grad_id, None)
        if cblist == None:
            cblist = [callback]
            self.callback_dict[grad_id] = cblist
            return
        cblist.append(callback)

    def update(self, *args, **kwargs):
        raise NotImplementedError('update not done for GradientContainer')

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)
        callbacks = self.callback_dict.get(key, [])
        for callback in callbacks:
            callback(val)


class Gradient(object):
    def __init__(self, element, svg):
        self.element = element
        self.stops = {}
        for e in element:
            if e.tag.endswith('stop'):
                style = parse_style(e.get('style', ''))
                color = parse_color(e.get('stop-color'), svg.current_color)
                if 'stop-color' in style:
                    color = parse_color(style['stop-color'], svg.current_color)
                color[3] = int(float(e.get('stop-opacity', '1')) * 255)
                if 'stop-opacity' in style:
                    color[3] = int(float(style['stop-opacity']) * 255)
                self.stops[float(e.get('offset'))] = color
        self.stops = sorted(self.stops.items())
        self.svg = svg
        self.inv_transform = Matrix(element.get('gradientTransform')).inverse()

        inherit = self.element.get('{http://www.w3.org/1999/xlink}href')
        parent = None
        delay_params = False
        if inherit:
            parent_id = inherit[1:]
            parent = self.svg.gradients.get(parent_id, None)
            if parent == None:
                self.svg.gradients.call_me_on_add(self.tardy_gradient_parsed, parent_id)
                delay_params = True
                return
        if not delay_params:
            self.get_params(parent)

    def interp(self, float x, float y):
        cdef Matrix m = self.inv_transform
        if not self.stops:
            return [255, 0, 255, 255]
        m.transform(x, y, &x, &y)
        t = self.grad_value(x, y)
        if t < self.stops[0][0]:
            return self.stops[0][1]
        for n, top in enumerate(self.stops[1:]):
            bottom = self.stops[n]
            if t <= top[0]:
                u = bottom[0]
                v = top[0]
                alpha = (t - u)/(v - u)
                return [int(item[0] * (1 - alpha) + item[1] * alpha) for item in zip(bottom[1], top[1])]
        return self.stops[-1][1]

    def get_params(self, parent):
        for param in self.params:
            v = None
            if parent:
                v = getattr(parent, param, None)
            my_v = self.element.get(param)
            if my_v:
                v = <float>float(my_v)
            if v:
                setattr(self, param, v)

    def tardy_gradient_parsed(self, gradient):
        self.get_params(gradient)


class LinearGradient(Gradient):
    params = ['x1', 'x2', 'y1', 'y2', 'stops']

    def grad_value(self, x, y):
        return ((x - self.x1)*(self.x2 - self.x1) + (y - self.y1)*(self.y2 - self.y1)) / ((self.x1 - self.x2)**2 + (self.y1 - self.y2)**2)


class RadialGradient(Gradient):
    params = ['cx', 'cy', 'r', 'stops']

    def grad_value(self, x, y):
        return sqrt((x - self.cx) ** 2 + (y - self.cy) ** 2)/self.r


cdef class Svg(RenderContext):
    """Svg class. See module for more information about the usage.
    """

    def __init__(self, source=None, anchor_x=0, anchor_y=0,
                 bezier_points=BEZIER_POINTS, circle_points=CIRCLE_POINTS,
                 color=None):
        '''
        Creates an SVG object from a .svg or .svgz file.

        :param str source: The name of the file to be loaded.
        :param float anchor_x: The horizontal anchor position for scaling and
            rotations. Defaults to 0. The symbolic values 'left', 'center' and
            'right' are also accepted.
        :param float anchor_y: The vertical anchor position for scaling and
            rotations. Defaults to 0. The symbolic values 'bottom', 'center' and
            'top' are also accepted.
        :param int bezier_points: The number of line segments into which to
            subdivide Bezier splines. Defaults to 10.
        :param int circle_points: The number of line segments into which to
            subdivide circular and elliptic arcs. Defaults to 10.
        :param color the default color to use for Svg elements that specify "currentColor"

        .. note:: if you want to use SVGs from string, you can parse the source yourself
            using `from xml.etree.cElementTree import fromstring` and pass the result to
            Svg().set_tree(). This will trigger the rendering of the Svg - as an alternative
            to assigning a filepath to Svg.source. This is also viable to trigger reloading.

        .. versionchanged:: 2.0.0
            Parameter `filename` changed to `source` and made optional.
        '''

        super(Svg, self).__init__(fs=SVG_FS, vs=SVG_VS,
                use_parent_projection=True,
                use_parent_modelview=True)

        self.last_mesh = None
        self.paths = []
        self.width = 0
        self.height = 0
        self.line_width = 1.0
        self.vbox_x = 0.
        self.vbox_y = 0.
        self.vbox_width = 0.
        self.vbox_height = 0.

        # if color is None:
        #     self.current_color = [0, 0, 0, 255]
        # else:
        #     self.current_color = kv_color_to_int_color(color)

        self.bezier_points = bezier_points
        self.circle_points = circle_points
        self.bezier_coefficients = None
        self.gradients = GradientContainer()
        self.anchor_x = anchor_x
        self.anchor_y = anchor_y
        self.line_texture = Texture.create(
                size=(2, 1), colorfmt="rgba")
        self.line_texture.blit_buffer(
                b"\xff\xff\xff\xff\xff\xff\xff\x00", colorfmt="rgba")

        self._source = None
        if source:
            self.source = source


    @property
    def anchor_x(self):
        '''
        Horizontal anchor position for scaling and rotations. Defaults to 0. The
        symbolic values 'left', 'center' and 'right' are also accepted.
        '''
        return self._anchor_x

    @anchor_x.setter
    def anchor_x(self, anchor_x):
        self._anchor_x = anchor_x
        if self._anchor_x == 'left':
            self._a_x = 0
        elif self._anchor_x == 'center':
            self._a_x = self.width * .5
        elif self._anchor_x == 'right':
            self._a_x = self.width
        else:
            self._a_x = self._anchor_x

    @property
    def anchor_y(self):
        '''
        Vertical anchor position for scaling and rotations. Defaults to 0. The
        symbolic values 'bottom', 'center' and 'top' are also accepted.
        '''
        return self._anchor_y

    @anchor_y.setter
    def anchor_y(self, anchor_y):
        self._anchor_y = anchor_y
        if self._anchor_y == 'bottom':
            self._a_y = 0
        elif self._anchor_y == 'center':
            self._a_y = self.height * .5
        elif self._anchor_y == 'top':
            self._a_y = self.height
        else:
            self._a_y = self.anchor_y

    @property
    def color(self):
        '''The default color

        Used for SvgElements that specify "currentColor"

        .. versionchanged:: 1.10.3

            The color is gettable and settable

        .. versionadded:: 1.9.1
        '''
        return int_color_to_kv_color(self.current_color)

    @color.setter
    def color(self, color):
        self.current_color = kv_color_to_int_color(color)
        self.reload()

    @property
    def source(self):
        '''Filename / source to load.

        The parsing and rendering is done as soon as you set the source.

        .. versionchanged:: 2.0.0
            The property name is now `source` instead of `filename`

        .. versionchanged:: 1.10.3
            You can get the used filename
        '''
        return self._source

    @source.setter
    def source(self, filename):
        Logger.debug('Svg: Loading {}'.format(filename))
        # check gzip
        start = time()
        with open(filename, 'rb') as fd:
            header = fd.read(3)
        if header == '\x1f\x8b\x08':
            import gzip
            fd = gzip.open(filename, 'rb')
        else:
            fd = open(filename, 'rb')
        try:
            #save the tree for later reloading
            self.set_tree(parse(fd))
            end = time()
            Logger.debug("Svg: Loaded {} in {:.2f}s".format(filename, end - start))
        finally:
            self._source = filename
            fd.close()

    def set_tree(self, tree):
        '''
        sets the tree used to render the Svg and triggers reloading.

        :param xml.etree.cElementTree tree: the tree parsed from the SVG source

        .. versionadded:: 2.0.0
        '''
        self.tree = tree
        self.reload()

    cdef void reload(self) except *:
            # parse tree
            start = time()
            self.parse_tree(self.tree)
            end1 = time()
            with self:
                self.render()
            end2 = time()
            Logger.debug("Svg: Parsed in {:.2f}s, rendered in {:.2f}s".format(
                    end1 - start, end2 - end1))

    cdef parse_tree(self, tree):
        root = tree._root
        self.paths = []
        self.width = parse_float(root.get('width'))
        self.height = parse_float(root.get('height'))

        view_box = parse_list(root.get('viewBox', '0 0 100% 100%'))
        self.vbox_x = parse_float(view_box[0])
        self.vbox_y = parse_float(view_box[1])
        self.vbox_width = parse_width(view_box[2], Window.width)
        self.vbox_height = parse_height(view_box[3], Window.height)

        if self.height:
            self.transform = Matrix([1, 0, 0, -1, 0, self.height])
        else:
            # XXX parse_width/height
            x, y, w, h = [parse_float(x) for x in
                    parse_list(root.get('viewBox'))]
            self.transform = Matrix([1, 0, 0, -1, -x, h + y])
            self.height = h
            self.width = w

        self.opacity = 1.0
        for e in root:
            self.parse_element(e)

    cdef parse_element(self, e):
        self.fill = parse_color(e.get('fill', 'black'), self.current_color)
        self.stroke = parse_color(e.get('stroke'), self.current_color)
        oldopacity = self.opacity
        self.opacity *= <float>float(e.get('opacity', 1))
        fill_opacity = <float>float(e.get('fill-opacity', 1))
        stroke_opacity = <float>float(e.get('stroke-opacity', 1))
        old_line_width = self.line_width
        self.line_width = <float>float(e.get('stroke-width', self.line_width))

        oldtransform = self.transform
        for t in self.parse_transform(e.get('transform')):
            self.transform *= Matrix(t)

        style = e.get('style')
        if style:
            sdict = parse_style(style)
            if 'fill' in sdict:
                self.fill = parse_color(sdict['fill'], self.current_color)
            if 'fill-opacity' in sdict:
                fill_opacity *= <float>float(sdict['fill-opacity'])
            if 'stroke' in sdict:
                self.stroke = parse_color(sdict['stroke'], self.current_color)
            if 'stroke-opacity' in sdict:
                stroke_opacity *= <float>float(sdict['stroke-opacity'])
            if 'stroke-width' in sdict:
                self.line_width = parse_float(sdict['stroke-width'])

        if self.stroke is None:
            self.stroke = [0, 0, 0, 0]
        if isinstance(self.stroke, list):
            self.stroke[3] = int(self.opacity * stroke_opacity * self.stroke[3])
        if isinstance(self.fill, list):
            self.fill[3] = int(self.opacity * fill_opacity * self.fill[3])
        if isinstance(self.stroke, list) and self.stroke[3] == 0:
            self.stroke = self.fill #Stroked edges antialias better

        if e.tag.endswith('path'):
            self.parse_path(e.get('d', ''))

        elif e.tag.endswith('rect'):
            x = 0
            y = 0
            if 'x' in e.keys():
                x = parse_width(e.get('x'), self.vbox_width)
            if 'y' in e.keys():
                y = parse_height(e.get('y'), self.vbox_height)
            h = parse_width(e.get('height'), self.vbox_width)
            w = parse_height(e.get('width'), self.vbox_height)
            rx = parse_width(e.get('rx', '0'), self.vbox_width)
            ry = parse_height(e.get('ry', '0'), self.vbox_height)

            if rx:
                if not ry:
                    ry = rx

                rx = min(rx, w / 2.)
                ry = min(ry, h / 2.)

            self.new_path()
            self.set_position(x + rx, y)
            self.set_position(x + w - rx, y)
            # top-right angle
            if rx:
                self.arc_to(rx, ry, 0, 0, 1, x + w, y + ry)

            self.set_position(x + w, y + h - ry)
            # bottom-right angle
            if rx:
                self.arc_to(rx, ry, 0, 0, 1, x + w - rx, y + h)

            self.set_position(x + rx, y + h)
            # bottom-left angle
            if rx:
                self.arc_to(rx, ry, 0, 0, 1, x, y + h - ry)

            self.set_position(x, y + ry)
            # top-left angle
            if rx:
                self.arc_to(rx, ry, 0, 0, 1, x + rx, y)

            self.end_path()

        elif e.tag.endswith('polyline') or e.tag.endswith('polygon'):
            pathdata = e.get('points')
            pathdata = re.findall(RE_POLYLINE, pathdata)
            pathdata.reverse()

            self.new_path()
            while pathdata:
                self.set_position(
                    parse_width(pathdata.pop(), self.vbox_width),
                    parse_height(pathdata.pop(), self.vbox_height))
            if e.tag.endswith('polygon'):
                self.close_path()
            self.end_path()

        elif e.tag.endswith('line'):
            x1 = parse_height(e.get('x1'), self.vbox_width)
            y1 = parse_width(e.get('y1'), self.vbox_height)
            x2 = parse_height(e.get('x2'), self.vbox_width)
            y2 = parse_width(e.get('y2'), self.vbox_height)
            self.new_path()
            self.set_position(x1, y1)
            self.set_position(x2, y2)
            self.end_path()

        elif e.tag.endswith('circle'):
            cx = parse_width(e.get('cx'), self.vbox_width)
            cy = parse_height(e.get('cy'), self.vbox_height)
            r = parse_float(e.get('r'))
            self.new_path()
            for i in xrange(self.circle_points):
                theta = 2 * i * pi / self.circle_points
                self.set_position(cx + r * cos(theta), cy + r * sin(theta))
            self.close_path()
            self.end_path()

        elif e.tag.endswith('ellipse'):
            cx = parse_width(e.get('cx'), self.vbox_width)
            cy = parse_height(e.get('cy'), self.vbox_height)
            rx = parse_width(e.get('rx'), self.vbox_width)
            ry = parse_height(e.get('ry'), self.vbox_height)
            self.new_path()
            for i in xrange(self.circle_points):
                theta = 2 * i * pi / self.circle_points
                self.set_position(cx + rx * cos(theta), cy + ry * sin(theta))
            self.close_path()
            self.end_path()

        elif e.tag.endswith('linearGradient'):
            self.gradients[e.get('id')] = LinearGradient(e, self)

        elif e.tag.endswith('radialGradient'):
            self.gradients[e.get('id')] = RadialGradient(e, self)

        for c in e:
            self.parse_element(c)

        self.transform = oldtransform
        self.opacity = oldopacity
        self.line_width = old_line_width

    cdef list parse_transform(self, transform_def):
        if isinstance(transform_def, str):
            return RE_TRANSFORM.findall(transform_def)
        else:
            return [transform_def]

    cdef parse_path(self, pathdef):
        # In the SVG specs, initial movetos are absolute, even if
        # specified as 'm'. This is the default behavior here as well.
        # But if you pass in a current_pos variable, the initial moveto
        # will be relative to that current_pos. This is useful.
        elements = list(_tokenize_path(pathdef))
        # Reverse for easy use of .pop()
        elements.reverse()
        command = None

        self.new_path()

        while elements:

            if elements[-1] in COMMANDS:
                # New command.
                last_command = command # Used by S and T
                command = elements.pop()
                absolute = command in UPPERCASE
                command = command.upper()
            else:
                # If this element starts with numbers, it is an implicit command
                # and we don't change the command. Check that it's allowed:
                if command is None:
                    raise ValueError("Unallowed implicit command in %s, position %s" % (
                        pathdef, len(pathdef.split()) - len(elements)))

            if command == 'M':
                # Moveto command. This is like "picking up the pen", so
                # start a new loop.
                if len(self.loop):
                    self.path.append(self.loop)
                    self.loop = array('f', [])

                x = parse_width(elements.pop(), self.vbox_width)
                y = parse_height(elements.pop(), self.vbox_height)
                self.set_position(x, y, absolute)

                # Implicit moveto commands are treated as lineto commands.
                # So we set command to lineto here, in case there are
                # further implicit commands after this moveto.
                command = 'L'

            elif command == 'Z':
                self.close_path()

            elif command == 'L':
                x = parse_width(elements.pop(), self.vbox_width)
                y = parse_height(elements.pop(), self.vbox_height)
                self.set_position(x, y, absolute)

            elif command == 'H':
                x = parse_width(elements.pop(), self.vbox_width)
                if absolute:
                    self.set_position(x, self.y)
                else:
                    self.set_position(self.x + x, self.y)

            elif command == 'V':
                y = parse_height(elements.pop(), self.vbox_height)
                if absolute:
                    self.set_position(self.x, y)
                else:
                    self.set_position(self.x, self.y + y)

            elif command == 'C':
                c1x = parse_width(elements.pop(), self.vbox_width)
                c1y = parse_height(elements.pop(), self.vbox_height)
                c2x = parse_width(elements.pop(), self.vbox_width)
                c2y = parse_height(elements.pop(), self.vbox_height)
                endx = parse_width(elements.pop(), self.vbox_width)
                endy = parse_height(elements.pop(), self.vbox_height)

                if not absolute:
                    c1x += self.x
                    c1y += self.y
                    c2x += self.x
                    c2y += self.y
                    endx += self.x
                    endy += self.y

                self.curve_to(c1x, c1y, c2x, c2y, endx, endy)

            elif command == 'S':
                # Smooth curve. First control point is the "reflection" of
                # the second control point in the previous path.

                if last_command not in 'CS':
                    # If there is no previous command or if the previous command
                    # was not an C, c, S or s, assume the first control point is
                    # coincident with the current point.
                    c1x = self.x
                    c1y = self.y
                else:
                    # The first control point is assumed to be the reflection of
                    # the second control point on the previous command relative
                    # to the current point.
                    c1x = self.x + self.x - self.last_cx
                    c1y = self.y + self.y - self.last_cy

                c2x = parse_width(elements.pop(), self.vbox_width)
                c2y = parse_height(elements.pop(), self.vbox_height)
                endx = parse_width(elements.pop(), self.vbox_width)
                endy = parse_height(elements.pop(), self.vbox_height)

                if not absolute:
                    c2x += self.x
                    c2y += self.y
                    endx += self.x
                    endy += self.y

                self.curve_to(c1x, c1y, c2x, c2y, endx, endy)
                # if we have multiple sets of coords, we want to use the
                # last control point, not new position, as next control
                # point
                last_command = 'S'

            elif command == 'A':
                rx = <float>float(elements.pop())
                ry = <float>float(elements.pop())
                rotation = <float>float(elements.pop())
                arc = <float>float(elements.pop())
                sweep = <float>float(elements.pop())
                x = parse_width(elements.pop(), self.vbox_width)
                y = parse_height(elements.pop(), self.vbox_height)

                if not absolute:
                    x += self.x
                    y += self.y

                self.arc_to(rx, ry, rotation, arc, sweep, x, y)

            elif command == 'Q':
                cx = parse_width(elements.pop(), self.vbox_width)
                cy = parse_height(elements.pop(), self.vbox_height)
                x = parse_width(elements.pop(), self.vbox_width)
                y = parse_height(elements.pop(), self.vbox_height)

                if not absolute:
                    cx += self.x
                    cy += self.y
                    x += self.x
                    y += self.y

                self.quadratic_bezier_curve_to(cx, cy, x, y)

            elif command == 'T':
                # Smooth curve. Control point is the "reflection" of
                # the second control point in the previous path.

                if last_command not in 'QT':
                    # If there is no previous command or if the previous command
                    # was not an Q, q, T or t, assume the first control point is
                    # coincident with the current point.
                    cx = self.x
                    cy = self.y
                else:
                    # The control point is assumed to be the reflection of
                    # the control point on the previous command relative
                    # to the current point.
                    cx = self.x + self.x - cx
                    cy = self.y + self.y - cy

                x = parse_width(elements.pop(), self.vbox_width)
                y = parse_height(elements.pop(), self.vbox_height)

                if not absolute:
                    x += self.x
                    y += self.y

                self.quadratic_bezier_curve_to(cx, cy, x, y)

            else:
                Logger.warning('Svg: unimplemented command {}'.format(command))

        self.end_path()

    cdef void new_path(self):
        self.x = 0
        self.y = 0
        self.close_index = 0
        self.path = []
        self.loop = array('f', [])

    cdef void close_path(self):
        if len(self.loop):
            self.loop.append(self.loop[0])
            self.loop.append(self.loop[1])

    cdef void set_position(self, double x, double y, int absolute=1):
        if absolute:
            self.x = <float>x
            self.y = <float>y
        else:
            self.x += <float>x
            self.y += <float>y
        self.loop.append(self.x)
        self.loop.append(self.y)

    cdef arc_to(self, double rx, double ry, double phi, double large_arc,
            double sweep, double x, double y):
        # This function is made out of magical fairy dust
        # http://www.w3.org/TR/2003/REC-SVG11-20030114/implnote.html#ArcImplementationNotes
        cdef double x1, y1, x2, y2, cp, sp, dx, dy, x_, y_, r2, cx_, cy_, cx, cy
        cdef double psi, delta, ct, st, theta
        cdef int n_points, i
        x1 = self.x
        y1 = self.y
        x2 = x
        y2 = y
        cp = cos(phi)
        sp = sin(phi)
        dx = .5 * (x1 - x2)
        dy = .5 * (y1 - y2)
        x_ = cp * dx + sp * dy
        y_ = -sp * dx + cp * dy
        r2 = (((rx * ry)**2 - (rx * y_)**2 - (ry * x_)**2)/
          ((rx * y_)**2 + (ry * x_)**2))
        if r2 < 0: r2 = 0
        r = sqrt(r2)
        if large_arc == sweep:
            r = -r
        cx_ = r * rx * y_ / ry
        cy_ = -r * ry * x_ / rx
        cx = cp * cx_ - sp * cy_ + .5 * (x1 + x2)
        cy = sp * cx_ + cp * cy_ + .5 * (y1 + y2)

        psi = angle(1, 0, (x_ - cx_) / rx, (y_ - cy_) / ry)
        delta = angle((x_ - cx_) / rx, (y_ - cy_) / ry,
                      (-x_ - cx_) / rx, (-y_ - cy_) / ry)
        if sweep and delta < 0: delta += pi * 2
        if not sweep and delta > 0: delta -= pi * 2
        n_points = <int>fabs(self.circle_points * delta / (2 * pi))
        if n_points < 1:
            n_points = 1

        for i in range(n_points + 1):
            theta = psi + i * delta / n_points
            ct = cos(theta)
            st = sin(theta)
            self.set_position(cp * rx * ct - sp * ry * st + cx,
                    sp * rx * ct + cp * ry * st + cy)


    @cython.boundscheck(False)
    cdef void quadratic_bezier_curve_to(self, float cx, float cy, float x, float y):
        cdef int bp_count = self.bezier_points + 1
        cdef int i, count, ilast
        cdef float t, t0, t1, t2, px = 0, py = 0
        cdef list bc
        cdef array.array loop
        cdef float* f_loop
        cdef float[:] f_bc

        if self.bezier_coefficients is None:
            self.bezier_coefficients = view.array(
                    shape=(bp_count * 3, ),
                    itemsize=sizeof(float),
                    format="f")
            f_bc = self.bezier_coefficients
            for i in range(bp_count):
                t = <float>(i / self.bezier_points)
                t0 = <float>pow(1 - t, 2)
                t1 = <float>(2 * t * (1 - t))
                t2 = <float>pow(t, 2)
                f_bc[i * 3] = t0
                f_bc[i * 3 + 1] = t1
                f_bc[i * 3 + 2] = t2
        else:
            f_bc = self.bezier_coefficients

        self.last_cx = cx
        self.last_cy = cy
        count = bp_count * 2
        ilast = <int>len(self.loop)
        array.resize(self.loop, ilast + count)
        f_loop = self.loop.data.as_floats
        for i in range(bp_count):
            t0 = f_bc[i * 3]
            t1 = f_bc[i * 3 + 1]
            t2 = f_bc[i * 3 + 2]
            f_loop[ilast + i * 2] = px = t0 * self.x + t1 * cx + t2 * x
            f_loop[ilast + i * 2 + 1] = py = t0 * self.y + t1 * cy + t2 * y

        self.x, self.y = px, py

    @cython.boundscheck(False)
    cdef void curve_to(self, float x1, float y1, float x2, float y2,
            float x, float y):
        cdef int bp_count = self.bezier_points + 1
        cdef int i, count, ilast
        cdef float t, t0, t1, t2, t3, px = 0, py = 0
        cdef list bc
        cdef array.array loop
        cdef float* f_loop
        cdef float[:] f_bc

        if self.bezier_coefficients is None:
            self.bezier_coefficients = view.array(
                    shape=(bp_count * 4, ),
                    itemsize=sizeof(float),
                    format="f")
            f_bc = self.bezier_coefficients
            for i in range(bp_count):
                t = <float>float(i) / self.bezier_points
                t0 = <float>pow(1 - t, 3)
                t1 = <float>(3 * t * pow(1 - t, 2))
                t2 = <float>(3 * pow(t, 2) * (1 - t))
                t3 = <float>pow(t, 3)
                f_bc[i * 4] = t0
                f_bc[i * 4 + 1] = t1
                f_bc[i * 4 + 2] = t2
                f_bc[i * 4 + 3] = t3
        else:
            f_bc = self.bezier_coefficients

        self.last_cx = x2
        self.last_cy = y2
        count = bp_count * 2
        ilast = <int>len(self.loop)
        array.resize(self.loop, ilast + count)
        f_loop = self.loop.data.as_floats
        for i in range(bp_count):
            t0 = f_bc[i * 4]
            t1 = f_bc[i * 4 + 1]
            t2 = f_bc[i * 4 + 2]
            t3 = f_bc[i * 4 + 3]
            f_loop[ilast + i * 2] = px = t0 * self.x + t1 * x1 + t2 * x2 + t3 * x
            f_loop[ilast + i * 2 + 1] = py = t0 * self.y + t1 * y1 + t2 * y2 + t3 * y

        self.x, self.y = px, py

    cdef void end_path(self):
        if len(self.loop):
            self.path.append(self.loop)
        tris = None
        cdef Tesselator tess
        cdef array.array loop
        if self.fill:
            tess = Tesselator()
            for loop in self.path:
                tess.add_contour_data(loop.data.as_voidptr, <int>int(len(loop) / 2.))
            tess.tesselate()
            tris = tess.vertices

        # Add the stroke for the first subpath, and the fill for all
        # subpaths.
        for loop in self.path:
            if self.stroke:
                self.paths.append((
                    loop,
                    self.stroke,
                    tris,
                    self.fill,
                    self.transform,
                    self.line_width))

            if self.fill:
                self.paths.append((
                    loop,
                    self.stroke,
                    None,
                    None,
                    self.transform,
                    0.))
        self.path = []

    @cython.boundscheck(False)
    cdef void push_mesh(self, float[:] path, fill, Matrix transform, mode):
        cdef float *vertices
        cdef int index, vindex
        cdef float *f_tris
        cdef float x, y, r, g, b, a
        cdef Mesh mesh

        cdef int count = <int>int(len(path) / 2.)
        vertices = <float *>malloc(sizeof(float) * count * 8)
        if vertices == NULL:
            return
        vindex = 0

        if isinstance(fill, str):
            gradient = self.gradients[fill]
            for index in range(count):
                x = path[index * 2]
                y = path[index * 2 + 1]
                r, g, b, a = gradient.interp(x, y)
                transform.transform(x, y, &x, &y)
                vertices[vindex] = x
                vertices[vindex + 1] = y
                vertices[vindex + 2] = 0
                vertices[vindex + 3] = 0
                vertices[vindex + 4] = r
                vertices[vindex + 5] = g
                vertices[vindex + 6] = b
                vertices[vindex + 7] = a
                vindex += 8
        else:
            r, g, b, a = fill
            for index in range(count):
                x = path[index * 2]
                y = path[index * 2 + 1]
                transform.transform(x, y, &x, &y)
                vertices[vindex] = x
                vertices[vindex + 1] = y
                vertices[vindex + 2] = 0
                vertices[vindex + 3] = 0
                vertices[vindex + 4] = r
                vertices[vindex + 5] = g
                vertices[vindex + 6] = b
                vertices[vindex + 7] = a
                vindex += 8

        self.push_strip_mesh(vertices, vindex, count)
        free(vertices)

    cdef void push_strip_mesh(self, float *vertices, int vindex, int count,
            int mode=0):
        if self.last_mesh:
            if self.last_mesh.add_triangle_strip(vertices, vindex, count, mode):
                return
        self.last_mesh = StripMesh(fmt=VERTEX_FORMAT)
        self.last_mesh.add_triangle_strip(vertices, vindex, count, mode)

    cdef void push_line_mesh(self, float[:] path, fill, Matrix transform, float width):
        # Tentative to use smooth line, doesn't work completely yet.
        # Caps and joint are missing
        cdef int index, vindex = 0, odd = 0, i
        cdef float ax, ay, bx, _by, r = 0, g = 0, b = 0, a = 0
        cdef int count = <int>int(len(path) / 2.)
        cdef float *vertices = NULL
        vindex = 0

        vertices = <float *>malloc(sizeof(float) * count * 32)
        if vertices == NULL:
            return

        if not isinstance(fill, str):
            r, g, b, a = fill

        for index in range(count - 1):
            i = index * 2
            ax = path[i]
            ay = path[i + 1]
            if index == count - 1:
                i = 0
            else:
                i = index * 2 + 2
            bx = path[i]
            _by = path[i + 1]
            transform.transform(ax, ay, &ax, &ay)
            transform.transform(bx, _by, &bx, &_by)

            rx = bx - ax
            ry = _by - ay
            angle = atan2(ry, rx)
            a1 = angle - PI2
            a2 = angle + PI2

            cos1 = cos(a1) * width
            sin1 = sin(a1) * width
            cos2 = cos(a2) * width
            sin2 = sin(a2) * width

            x1 = ax + cos1
            y1 = ay + sin1
            x4 = ax + cos2
            y4 = ay + sin2
            x2 = bx + cos1
            y2 = _by + sin1
            x3 = bx + cos2
            y3 = _by + sin2

            if isinstance(fill, str):
                g = self.gradients[fill]
                r, g, b, a = g.interp(ax, ay)

            vertices[vindex + 2] = vertices[vindex + 10] = \
                vertices[vindex + 18] = vertices[vindex + 26] = 0
            vertices[vindex + 3] = vertices[vindex + 11] = \
                vertices[vindex + 19] = vertices[vindex + 27] = 0
            vertices[vindex + 4] = vertices[vindex + 12] = \
                vertices[vindex + 20] = vertices[vindex + 28] = r
            vertices[vindex + 5] = vertices[vindex + 13] = \
                vertices[vindex + 21] = vertices[vindex + 29] = g
            vertices[vindex + 6] = vertices[vindex + 14] = \
                vertices[vindex + 22] = vertices[vindex + 30] = b
            vertices[vindex + 7] = vertices[vindex + 15] = \
                vertices[vindex + 23] = vertices[vindex + 31] = a

            vertices[vindex + 0] = <float>x1
            vertices[vindex + 1] = <float>y1
            vertices[vindex + 8] = <float>x4
            vertices[vindex + 9] = <float>y4
            vertices[vindex + 16] = <float>x2
            vertices[vindex + 17] = <float>y2
            vertices[vindex + 24] = <float>x3
            vertices[vindex + 25] = <float>y3
            vindex += 32

        # if self.closed:
        #     vindex = vcount - 4
        #     i0 = vindex
        #     i1 = vindex + 1
        #     i2 = vindex + 2
        #     i3 = vindex + 3
        #     i4 = 0
        #     i5 = 1
        #     i6 = 2
        #     i7 = 3
        #     tindices[0] = i0
        #     tindices[1] = i2
        #     tindices[2] = i6
        #     tindices[3] = i0
        #     tindices[4] = i6
        #     tindices[5] = i4
        #     tindices[6] = i1
        #     tindices[7] = i0
        #     tindices[8] = i4
        #     tindices[9] = i1
        #     tindices[10] = i4
        #     tindices[11] = i5
        #     tindices[12] = i3
        #     tindices[13] = i1
        #     tindices[14] = i5
        #     tindices[15] = i3
        #     tindices[16] = i5
        #     tindices[17] = i7
        #     tindices = tindices + 18

        self.push_strip_mesh(vertices, vindex, <int>int((vindex / 32.)) * 4, 1)
        free(vertices)

    cdef void render(self):
        for path, stroke, tris, fill, transform, width in self.paths:
            if tris:
                for item in tris:
                    self.push_mesh(item, fill, transform, 'triangle_strip')
            if path:
                self.push_line_mesh(path, stroke, transform, width)



'''
Stencil instructions
====================

.. versionadded:: 1.0.4

.. versionchanged:: 1.3.0
    The stencil operation has been updated to resolve some issues that appeared
    when nested. You **must** now have a StencilUnUse and repeat the same
    operation as you did after StencilPush.

Stencil instructions permit you to draw and use the current drawing as a mask.
They don't give as much control as pure OpenGL, but you can still do fancy
things!

The stencil buffer can be controlled using these 3 instructions:

    - :class:`StencilPush`: push a new stencil layer.
      Any drawing that happens after this will be used as a mask.
    - :class:`StencilUse` : now draw the next instructions and use the stencil
      for masking them.
    - :class:`StencilUnUse` : stop using the stencil i.e. remove the mask and
      draw normally.
    - :class:`StencilPop` : pop the current stencil layer.


You should always respect this scheme:

.. code-block:: kv

    StencilPush

    # PHASE 1: put any drawing instructions to use as a mask here.

    StencilUse

    # PHASE 2: all the drawing here will be automatically clipped by the
    # mask created in PHASE 1.

    StencilUnUse

    # PHASE 3: put the same drawing instruction here as you did in PHASE 1

    StencilPop

    # PHASE 4: the stencil is now removed from the stack and unloaded.


Limitations
-----------

- Drawing in PHASE 1 and PHASE 3 must not collide or you
  will get unexpected results
- The stencil is activated as soon as you perform a StencilPush
- The stencil is deactivated as soon as you've correctly popped all the stencil
  layers
- You must not play with stencils yourself between a StencilPush / StencilPop
- You can push another stencil after a StencilUse / before the StencilPop
- You can push up to 128 layers of stencils (8 for kivy < 1.3.0)


Example of stencil usage
------------------------

Here is an example, in kv style::

    StencilPush

    # create a rectangular mask with a pos of (100, 100) and a (100, 100) size.
    Rectangle:
        pos: 100, 100
        size: 100, 100

    StencilUse

    # we want to show a big green rectangle, however, the previous stencil
    # mask will crop us :)
    Color:
        rgb: 0, 1, 0
    Rectangle:
        size: 900, 900

    StencilUnUse

    # you must redraw the stencil mask to remove it
    Rectangle:
        pos: 100, 100
        size: 100, 100

    StencilPop

'''

__all__ = ('StencilPush', 'StencilPop', 'StencilUse', 'StencilUnUse')

include "../include/config.pxi"
include "opcodes.pxi"
include "gl_debug_logger.pxi"

from kivy.graphics.cgl cimport *
from kivy.compat import PY2
from kivy.graphics.instructions cimport Instruction

cdef dict DEFAULT_STATE = {
    "level": 0,
    "in_push": False,
    "op": None,
    "gl_stencil_func": None,
    "clear_stencil": True
}

cdef dict _stencil_state = DEFAULT_STATE.copy()

cdef dict _gl_stencil_op = {
    'never': GL_NEVER, 'less': GL_LESS, 'equal': GL_EQUAL,
    'lequal': GL_LEQUAL, 'greater': GL_GREATER, 'notequal': GL_NOTEQUAL,
    'gequal': GL_GEQUAL, 'always': GL_ALWAYS }


cdef inline int _stencil_op_to_gl(x):
    '''Return the GL numeric value from a stencil operator
    '''
    x = x.lower()
    try:
        return _gl_stencil_op[x]
    except KeyError:
        raise Exception('Unknown <%s> stencil op' % x)


cdef get_stencil_state():
    global _stencil_state
    return _stencil_state.copy()


cdef void restore_stencil_state(dict state):
    global _stencil_state
    _stencil_state = state.copy()
    stencil_apply_state(_stencil_state, True)


cdef void reset_stencil_state():
    restore_stencil_state(DEFAULT_STATE)


cdef void stencil_apply_state(dict state, restore_only):
    # apply state for stencil here. This allow to reapply a state
    # easily when using FBO, or linking to other GL subprogram
    if state["op"] is None:
        cgl.glDisable(GL_STENCIL_TEST)

    elif state["op"] == "push":
        # Push the stencil stack, ready to draw a mask

        if not restore_only:
            state["level"] += 1
            state["in_push"] = True

        if state["level"] == 1:
            cgl.glStencilMask(0xff)
            log_gl_error('StencilPush.apply-glStencilMask')
            if state["clear_stencil"]:
                cgl.glClearStencil(0)
                log_gl_error('StencilPush.apply-glClearStencil')
                cgl.glClear(GL_STENCIL_BUFFER_BIT)
                log_gl_error('StencilPush.apply-glClear(GL_STENCIL_BUFFER_BIT)')
        elif state["level"] > 128:
            raise Exception('Cannot push more than 128 level of stencil.'
                            ' (stack overflow)')

        cgl.glEnable(GL_STENCIL_TEST)
        log_gl_error('StencilPush.apply-glEnable(GL_STENCIL_TEST)')
        cgl.glStencilFunc(GL_ALWAYS, 1, 0xff)
        log_gl_error('StencilPush.apply-glStencilFunc')
        cgl.glStencilOp(GL_INCR, GL_INCR, GL_INCR)
        log_gl_error('StencilPush.apply-glStencilOp')
        cgl.glColorMask(False, False, False, False)
        log_gl_error('StencilPush.apply-glColorMask')


    elif state["op"] == "pop":
        # Pop the stencil stack

        if not restore_only:
            if state["level"] == 0:
                raise Exception('Too much StencilPop (stack underflow)')
            state["level"] -= 1
            state["in_push"] = False

        cgl.glColorMask(True, True, True, True)
        log_gl_error('StencilPop.apply-glColorMask')
        if state["level"] == 0:
            cgl.glDisable(GL_STENCIL_TEST)
            log_gl_error('StencilPop.apply-glDisable')
            return
        # reset for previous
        cgl.glStencilFunc(GL_EQUAL, state["level"], 0xff)
        log_gl_error('StencilPop.apply-glStencilFunc')
        cgl.glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP)
        log_gl_error('StencilPop.apply-glStencilOp')


    elif state["op"] == "use":
        # Use the current stencil buffer to cut the drawing
        if not restore_only:
            state["in_push"] = False
        cgl.glColorMask(True, True, True, True)
        log_gl_error('StencilUse.apply-glColorMask')
        cgl.glStencilFunc(state["gl_stencil_func"], state["level"], 0xff)
        log_gl_error('StencilUse.apply-glStencilFunc')
        cgl.glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP)
        cgl.glEnable(GL_STENCIL_TEST)
        log_gl_error('StencilUse.apply-glStencilOp')

    elif state["op"] == "unuse":
        # Ready to undraw the mask
        cgl.glStencilFunc(GL_GREATER, 0xff, 0xff)
        log_gl_error('StencilUnUse.apply-glStencilFunc')
        cgl.glStencilOp(GL_DECR, GL_DECR, GL_DECR)
        log_gl_error('StencilUnUse.apply-glStencilOp')
        cgl.glColorMask(False, False, False, False)
        log_gl_error('StencilUnUse.apply-glColorMask')


cdef class StencilPush(Instruction):
    '''Push the stencil stack. See the module documentation for more
    information.

    '''

    def __init__(self, **kwargs):
        super(StencilPush, self).__init__(**kwargs)
        self._clear_stencil = self._check_bool(kwargs.get('clear_stencil', True))

    cdef bint _check_bool(self, object value):
        if not isinstance(value, bool):
            raise TypeError(
                f"'clear_stencil' accept only boolean values (True or False), got {type(value)}."
            )
        return value

    cdef int apply(self) except -1:
        _stencil_state["op"] = "push"
        _stencil_state["clear_stencil"] = self._clear_stencil
        stencil_apply_state(_stencil_state, False)
        return 0

    @property
    def clear_stencil(self):
        '''``clear_stencil`` allow to disable stencil clearing in the ``StencilPush``
        phase. This option essentially disables the invocation of the functions
        ``cgl.glClearStencil(0)`` and ``cgl.glClear(GL_STENCIL_BUFFER_BIT).``

        If ``True``, the stencil will be cleaned in the ``StencilPush`` phase, if
        ``False``, it will not be cleaned.

        .. note::
            It is **highly recommended** to set ``clear_stencil=False`` for improved
            performance and reduced GPU usage (especially if there are hundreds of
            instructions). However, if any side effects (such as artifacts or inaccurate
            behavior of ``StencilPush``) occur, it is advisable to re-enable the clearing
            instructions with ``clear_stencil=True.``

        .. versionadded:: 2.3.0
        '''
        return self._clear_stencil

    @clear_stencil.setter
    def clear_stencil(self, value):
        cdef int clear_stencil = self._check_bool(value)
        if clear_stencil != self._clear_stencil:
            self._clear_stencil = clear_stencil
            self.flag_data_update()


cdef class StencilPop(Instruction):
    '''Pop the stencil stack. See the module documentation for more information.
    '''
    cdef int apply(self) except -1:
        _stencil_state["op"] = "pop"
        stencil_apply_state(_stencil_state, False)
        return 0


cdef class StencilUse(Instruction):
    '''Use current stencil buffer as a mask. Check the module documentation for
    more information.
    '''
    def __init__(self, **kwargs):
        super(StencilUse, self).__init__(**kwargs)
        if 'op' in kwargs:
            self._op = _stencil_op_to_gl(kwargs['op'])
        else:
            self._op = GL_EQUAL

    cdef int apply(self) except -1:
        _stencil_state["gl_stencil_func"] = self._op
        _stencil_state["op"] = "use"
        stencil_apply_state(_stencil_state, False)
        return 0

    @property
    def func_op(self):
        '''Determine the stencil operation to use for glStencilFunc(). Can be
        one of 'never', 'less', 'equal', 'lequal', 'greater', 'notequal',
        'gequal' or 'always'.

        By default, the operator is set to 'equal'.

        .. versionadded:: 1.5.0
        '''

        index = _gl_stencil_op.values().index(self._op)
        if PY2:
            return _gl_stencil_op.keys()[index]
        else:
            return list(_gl_stencil_op.keys())[index]

    @func_op.setter
    def func_op(self, x):
        cdef int op = _stencil_op_to_gl(x)
        if op != self._op:
            self._op = op
            self.flag_data_update()


cdef class StencilUnUse(Instruction):
    '''Use current stencil buffer to unset the mask.
    '''
    cdef int apply(self) except -1:
        _stencil_state["op"] = "unuse"
        stencil_apply_state(_stencil_state, False)
        return 0
